#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: pnp_global_settings
short_description: Resource module for Pnp Global Settings
description:
- Manage operation update of the resource Pnp Global Settings.
- Updates the user's list of global PnP settings.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  _id:
    description: Pnp Global Settings's _id.
    type: str
  aaaCredentials:
    description: Pnp Global Settings's aaaCredentials.
    suboptions:
      password:
        description: Pnp Global Settings's password.
        type: str
      username:
        description: Pnp Global Settings's username.
        type: str
    type: dict
  acceptEula:
    description: AcceptEula flag.
    type: bool
  defaultProfile:
    description: Pnp Global Settings's defaultProfile.
    suboptions:
      cert:
        description: Pnp Global Settings's cert.
        type: str
      fqdnAddresses:
        description: Pnp Global Settings's fqdnAddresses.
        elements: str
        type: list
      ipAddresses:
        description: Pnp Global Settings's ipAddresses.
        elements: str
        type: list
      port:
        description: Pnp Global Settings's port.
        type: int
      proxy:
        description: Proxy flag.
        type: bool
    type: dict
  savaMappingList:
    description: Pnp Global Settings's savaMappingList.
    elements: dict
    suboptions:
      autoSyncPeriod:
        description: Pnp Global Settings's autoSyncPeriod.
        type: int
      ccoUser:
        description: Pnp Global Settings's ccoUser.
        type: str
      expiry:
        description: Pnp Global Settings's expiry.
        type: int
      lastSync:
        description: Pnp Global Settings's lastSync.
        type: int
      profile:
        description: Pnp Global Settings's profile.
        suboptions:
          addressFqdn:
            description: Pnp Global Settings's addressFqdn.
            type: str
          addressIpV4:
            description: Pnp Global Settings's addressIpV4.
            type: str
          cert:
            description: Pnp Global Settings's cert.
            type: str
          makeDefault:
            description: MakeDefault flag.
            type: bool
          name:
            description: Pnp Global Settings's name.
            type: str
          port:
            description: Pnp Global Settings's port.
            type: int
          profileId:
            description: Pnp Global Settings's profileId.
            type: str
          proxy:
            description: Proxy flag.
            type: bool
        type: dict
      smartAccountId:
        description: Pnp Global Settings's smartAccountId.
        type: str
      syncResult:
        description: Pnp Global Settings's syncResult.
        suboptions:
          syncList:
            description: Pnp Global Settings's syncList.
            elements: dict
            suboptions:
              deviceSnList:
                description: Pnp Global Settings's deviceSnList.
                elements: str
                type: list
              syncType:
                description: Pnp Global Settings's syncType.
                type: str
            type: list
          syncMsg:
            description: Pnp Global Settings's syncMsg.
            type: str
        type: dict
      syncResultStr:
        description: Pnp Global Settings's syncResultStr.
        type: str
      syncStartTime:
        description: Pnp Global Settings's syncStartTime.
        type: int
      syncStatus:
        description: Pnp Global Settings's syncStatus.
        type: str
      tenantId:
        description: Pnp Global Settings's tenantId.
        type: str
      token:
        description: Pnp Global Settings's token.
        type: str
      virtualAccountId:
        description: Pnp Global Settings's virtualAccountId.
        type: str
    type: list
  taskTimeOuts:
    description: Pnp Global Settings's taskTimeOuts.
    suboptions:
      configTimeOut:
        description: Pnp Global Settings's configTimeOut.
        type: int
      generalTimeOut:
        description: Pnp Global Settings's generalTimeOut.
        type: int
      imageDownloadTimeOut:
        description: Pnp Global Settings's imageDownloadTimeOut.
        type: int
    type: dict
  tenantId:
    description: Pnp Global Settings's tenantId.
    type: str
  version:
    description: Pnp Global Settings's version.
    type: int
requirements:
- dnacentersdk >= 2.5.0
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Device Onboarding (PnP) UpdatePnPGlobalSettings
  description: Complete reference of the UpdatePnPGlobalSettings API.
  link: https://developer.cisco.com/docs/dna-center/#!update-pn-p-global-settings
notes:
  - SDK Method used are
    device_onboarding_pnp.DeviceOnboardingPnp.update_pnp_global_settings,

  - Paths used are
    put /dna/intent/api/v1/onboarding/pnp-settings,

"""

EXAMPLES = r"""
- name: Update all
  cisco.dnac.pnp_global_settings:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: present
    _id: string
    aaaCredentials:
      password: string
      username: string
    acceptEula: true
    defaultProfile:
      cert: string
      fqdnAddresses:
      - string
      ipAddresses:
      - string
      port: 0
      proxy: true
    savaMappingList:
    - autoSyncPeriod: 0
      ccoUser: string
      expiry: 0
      lastSync: 0
      profile:
        addressFqdn: string
        addressIpV4: string
        cert: string
        makeDefault: true
        name: string
        port: 0
        profileId: string
        proxy: true
      smartAccountId: string
      syncResult:
        syncList:
        - deviceSnList:
          - string
          syncType: string
        syncMsg: string
      syncResultStr: string
      syncStartTime: 0
      syncStatus: string
      tenantId: string
      token: string
      virtualAccountId: string
    taskTimeOuts:
      configTimeOut: 0
      generalTimeOut: 0
      imageDownloadTimeOut: 0
    tenantId: string
    version: 0

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "savaMappingList": [
        {
          "syncStatus": "string",
          "syncStartTime": 0,
          "syncResult": {
            "syncList": [
              {
                "syncType": "string",
                "deviceSnList": [
                  "string"
                ]
              }
            ],
            "syncMsg": "string"
          },
          "lastSync": 0,
          "tenantId": "string",
          "profile": {
            "port": 0,
            "addressIpV4": "string",
            "addressFqdn": "string",
            "profileId": "string",
            "proxy": true,
            "makeDefault": true,
            "cert": "string",
            "name": "string"
          },
          "token": "string",
          "expiry": 0,
          "ccoUser": "string",
          "smartAccountId": "string",
          "virtualAccountId": "string",
          "autoSyncPeriod": 0,
          "syncResultStr": "string"
        }
      ],
      "taskTimeOuts": {
        "imageDownloadTimeOut": 0,
        "configTimeOut": 0,
        "generalTimeOut": 0
      },
      "tenantId": "string",
      "aaaCredentials": {
        "password": "string",
        "username": "string"
      },
      "defaultProfile": {
        "fqdnAddresses": [
          "string"
        ],
        "proxy": true,
        "cert": "string",
        "ipAddresses": [
          "string"
        ],
        "port": 0
      },
      "acceptEula": true,
      "id": "string",
      "_id": "string",
      "version": 0
    }
"""



from __future__ import absolute_import, division, print_function

__metaclass__ = type

from ansible_collections.community.general.tests.unit.compat import unittest
from ansible_collections.community.general.tests.unit.compat.mock import MagicMock
from ansible_collections.community.digitalocean.plugins.modules.digital_ocean_kubernetes import (
    DOKubernetes,
)


class TestDOKubernetes(unittest.TestCase):
    def test_get_by_id_when_ok(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.rest = MagicMock()
        k.rest.get = MagicMock()
        k.rest.get.return_value.status_code = 200
        k.rest.get.return_value.json = {"foo": "bar"}
        self.assertEqual(k.get_by_id(), {"foo": "bar"})

    def test_get_by_id_when_not_ok(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.rest = MagicMock()
        k.rest.get = MagicMock()
        k.rest.get.return_value.status_code = 400
        k.rest.get.return_value.json = {"foo": "bar"}
        self.assertIsNone(k.get_by_id())

    def test_get_all_clusters_when_ok(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.rest = MagicMock()
        k.rest.get = MagicMock()
        k.rest.get.return_value.status_code = 200
        k.rest.get.return_value.json = {"foo": "bar"}
        self.assertEqual(k.get_all_clusters(), {"foo": "bar"})

    def test_get_all_clusters_when_not_ok(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.rest = MagicMock()
        k.rest.get = MagicMock()
        k.rest.get.return_value.status_code = 400
        k.rest.get.return_value.json = {"foo": "bar"}
        self.assertIsNone(k.get_all_clusters())

    def test_get_by_name_none(self):
        module = MagicMock()
        k = DOKubernetes(module)
        self.assertIsNone(k.get_by_name(None))

    def test_get_by_name_found(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.get_all_clusters = MagicMock()
        k.get_all_clusters.return_value = {"kubernetes_clusters": [{"name": "foo"}]}
        self.assertEqual(k.get_by_name("foo"), {"name": "foo"})

    def test_get_by_name_not_found(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.get_all_clusters = MagicMock()
        k.get_all_clusters.return_value = {"kubernetes_clusters": [{"name": "foo"}]}
        self.assertIsNone(k.get_by_name("foo2"))

    def test_get_kubernetes_kubeconfig_when_ok(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.rest = MagicMock()
        k.rest.get = MagicMock()
        k.rest.get.return_value.status_code = 200
        k.rest.get.return_value.body = "kubeconfig"
        self.assertEqual(k.get_kubernetes_kubeconfig(), "kubeconfig")

    def test_get_kubernetes_kubeconfig_when_not_ok(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.rest = MagicMock()
        k.rest.get = MagicMock()
        k.rest.get.return_value.status_code = 400
        k.rest.get.return_value.body = "kubeconfig"
        self.assertNotEqual(k.get_kubernetes_kubeconfig(), "kubeconfig")

    def test_get_kubernetes_when_found(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.get_by_name = MagicMock()
        k.get_by_name.return_value = {"id": 42}
        self.assertEqual(k.get_kubernetes(), {"id": 42})

    def test_get_kubernetes_when_not_found(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.get_by_name = MagicMock()
        k.get_by_name.return_value = None
        self.assertIsNone(k.get_kubernetes())

    def test_get_kubernetes_options_when_ok(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.rest = MagicMock()
        k.rest.get = MagicMock()
        k.rest.get.return_value.json = {"foo": "bar"}
        k.rest.get.return_value.status_code = 200
        self.assertEqual(k.get_kubernetes_options(), {"foo": "bar"})

    def test_get_kubernetes_options_when_not_ok(self):
        module = MagicMock()
        k = DOKubernetes(module)
        k.rest = MagicMock()
        k.rest.get = MagicMock()
        k.rest.get.return_value.json = {"foo": "bar"}
        k.rest.get.return_value.status_code = 400
        self.assertIsNone(k.get_kubernetes_options())

    def test_ensure_running_when_running(self):
        module = MagicMock()
        module.fail_json = MagicMock()

        k = DOKubernetes(module)
        k.end_time = 20
        k.wait_timeout = 1
        k.get_by_id = MagicMock()

        cluster = {"kubernetes_cluster": {"status": {"state": "running"}}}

        k.get_by_id.return_value = cluster

        time = MagicMock()
        time.time = MagicMock()
        time.time.return_value = 10
        time.sleep = MagicMock()

        self.assertEqual(k.ensure_running(), cluster)

    def test_ensure_running_when_not_running(self):
        module = MagicMock()
        module.fail_json = MagicMock()

        k = DOKubernetes(module)
        k.end_time = 20
        k.wait_timeout = -100
        k.get_by_id = MagicMock()

        cluster = {"kubernetes_cluster": {"status": {"state": "stopped"}}}

        k.get_by_id.return_value = cluster

        time = MagicMock()
        time.time = MagicMock()
        time.time.return_value = 20
        time.sleep = MagicMock()

        # module.fail_json.assert_called()
        assert True

    def test_create_ok(self):
        module = MagicMock()
        module.exit_json = MagicMock()
        module.fail_json = MagicMock()

        k = DOKubernetes(module)
        k.module = MagicMock()
        k.module.params = MagicMock()

        k.module.params.return_value = {"region": "nyc1"}

        k.get_kubernetes_options = MagicMock()

        kubernetes_options = {
            "options": {
                "regions": [{"name": "New York 1", "slug": "nyc1"}],
                "versions": [{"kubernetes_version": "1.18.8", "slug": "1.18.8-do.0"}],
                "sizes": [{"name": "s-1vcpu-2gb", "slug": "s-1vcpu-2gb"}],
            }
        }

        k.get_kubernetes_options.return_value = kubernetes_options

        k.get_kubernetes = MagicMock()
        k.get_kubernetes.return_value = {"foo": "bar"}
        k.cluster_id = MagicMock()
        k.cluster_id.return_value = 42

        k.rest = MagicMock()
        k.rest.post = MagicMock()
        k.rest.post.return_value.json = {"kubernetes_cluster": {"id": 42}}
        k.rest.post.return_value.status_code = 200
        k.ensure_running = MagicMock()
        k.cluster_id = MagicMock()
        k.module = MagicMock()

        k.create()
        k.module.exit_json.assert_called()

    def test_create_not_ok(self):
        module = MagicMock()
        module.exit_json = MagicMock()
        module.fail_json = MagicMock()

        k = DOKubernetes(module)
        k.module = MagicMock()
        k.module.params = MagicMock()

        k.module.params.return_value = {"region": "nyc1"}

        k.get_kubernetes_options = MagicMock()

        kubernetes_options = {
            "options": {
                "regions": [{"name": "New York 1", "slug": "nyc1"}],
                "versions": [{"kubernetes_version": "1.18.8", "slug": "1.18.8-do.0"}],
                "sizes": [{"name": "s-1vcpu-2gb", "slug": "s-1vcpu-2gb"}],
            }
        }

        k.get_kubernetes_options.return_value = kubernetes_options

        k.get_kubernetes = MagicMock()
        k.get_kubernetes.return_value = {"foo": "bar"}
        k.cluster_id = MagicMock()
        k.cluster_id.return_value = 42

        k.rest = MagicMock()
        k.rest.post = MagicMock()
        k.rest.post.return_value.json = {"kubernetes_cluster": {"id": 42}}
        k.rest.post.return_value.status_code = 400
        k.ensure_running = MagicMock()
        k.cluster_id = MagicMock()
        k.module = MagicMock()

        k.create()
        k.module.exit_json.assert_called()

    def test_delete_ok(self):
        module = MagicMock()
        module.exit_json = MagicMock()

        k = DOKubernetes(module)

        k.get_kubernetes = MagicMock()

        k.rest = MagicMock()
        k.rest.delete = MagicMock()
        k.rest.delete.return_value.id = 42
        k.rest.delete.return_value.status_code = 204

        k.delete()
        k.module.exit_json.assert_called()

    def test_delete_not_ok(self):
        module = MagicMock()
        module.exit_json = MagicMock()

        k = DOKubernetes(module)

        k.get_kubernetes = MagicMock()
        k.get_kubernetes.return_value = None

        k.delete()
        k.module.exit_json.assert_called()



# -*- coding: utf-8 -*-
# (c) 2018, Jordan Borean <jborean@redhat.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# Make coding more python3-ish
from __future__ import (absolute_import, division, print_function)
from sys import call_tracing
__metaclass__ = type

import ntpath
import os

from ansible.module_utils.common.text.converters import to_bytes
from ansible.playbook.task import Task

from ansible_collections.ansible.windows.tests.unit.compat.mock import MagicMock
from ansible_collections.ansible.windows.plugins.action import win_updates


# Some raw info on the updates in the test expectations
UPDATE_INFO = {
    '1abb2377-20ef-43ff-aabc-0de4711ab205': {
        'title': '2020-10 Security Update for Adobe Flash Player for Windows Server 2019 for x64-based Systems (KB4580325)',
        'kb': '4580325',
        'categories': ['Security Updates', 'Windows Server 2019'],
    },
    '89e11227-761b-4396-bf37-37a2b641fa84': {
        'title': '2021-01 Update for Windows Server 2019 for x64-based Systems (KB4589208)',
        'kb': '4589208',
        'categories': ['Updates', 'Windows Server 2019'],
    },
    '74819184-828b-4f97-bb4b-089a0c58e366': {
        'title': '2021-05 Cumulative Update for Windows Server 2019 (1809) for x64-based Systems (KB5003171)',
        'kb': '5003171',
        'categories': ['Security Updates'],
    },
    '9e94ae7a-f0f4-4ea1-b590-9f1388ee6126': {
        'title': '2021-05 Cumulative Update Preview for .NET Framework 3.5, 4.7.2 and 4.8 for Windows Server 2019 for x64 (KB5003396)',
        'kb': '5003396',
        'categories': ['Updates', 'Windows Server 2019'],
    },
    'a89e56f1-2002-40c6-a9af-3cf8be801df7': {
        'title': 'Security Intelligence Update for Microsoft Defender Antivirus - KB2267602 (Version 1.339.1923.0)',
        'kb': '2267602',
        'categories': ['Definition Updates', 'Microsoft Defender Antivirus'],
    },
    '5dae3a9d-4f41-445c-83f6-15fb4534e936': {
        'title': 'Security Intelligence Update for Microsoft Defender Antivirus - KB2267602 (Version 1.341.8.0)',
        'kb': '2267602',
        'categories': ['Definition Updates', 'Microsoft Defender Antivirus'],
    },
    '4326d7df-c256-4cca-99f7-c02a04443ec1': {
        'title': 'Security Intelligence Update for Microsoft Defender Antivirus - KB2267602 (Version 1.341.70.0)',
        'kb': '2267602',
        'categories': ['Definition Updates', 'Microsoft Defender Antivirus'],
    },
    '81929363-530d-4ccc-b9c7-8a1b89b20fe5': {
        'title': 'Security Intelligence Update for Microsoft Defender Antivirus - KB2267602 (Version 1.341.72.0)',
        'kb': '2267602',
        'categories': ['Definition Updates', 'Microsoft Defender Antivirus'],
    },
    '33a64099-ba99-4e7f-a2d7-cf7d7fc4029f': {
        'title': 'Security Update for Windows Server 2019 for x64-based Systems (KB4535680)',
        'kb': '4535680',
        'categories': ['Security Updates', 'Windows Server 2019'],
    },

    'f26a0046-1e1a-4305-8743-19c92c3095a5': {
        'title': 'Update for Removal of Adobe Flash Player for Windows Server 2019 for x64-based systems (KB4577586)',
        'kb': '4577586',
        'categories': ['Updates', 'Windows Server 2019'],
    },
    'd4919b6d-584d-436f-b877-dc3fc352a774': {
        'title': 'Windows Malicious Software Removal Tool x64 - v5.89 (KB890830)',
        'kb': '890830',
        'categories': ['Update Rollups', 'Windows Server 2016', 'Windows Server 2019'],
    },
}


class UpdateProgressHelper:
    """Mocks the functionality of POLL_SCRIPT to get update results for testing"""

    def __init__(self, test_name, default_rc=0, default_stderr=b''):
        self._path = os.path.abspath(os.path.join(__file__, '..', '..', '..', 'test_data', 'win_updates', test_name))
        self._reader = self._read_gen()
        self._rc = default_rc
        self._stderr = default_stderr

    def poll(self, cmd, **kwargs):
        return next(self._reader)

    def _read_gen(self):
        offset = 0
        lines = []
        with open(self._path, mode='rb') as fd:
            while True:
                line = fd.readline().strip()

                # Every blank line is treated as no more data is available and the script should return the values
                # retrieved. A subsequent call will continue to return the remaining data.
                if not line:
                    lines.append(to_bytes(offset))
                    b_stdout = b'\n'.join(lines)
                    lines = []
                    yield self._rc, b_stdout, self._stderr

                lines.append(line)
                offset += len(line)


def mock_connection_init(test_id, default_rc=0, default_stderr=b''):
    progress_helper = UpdateProgressHelper(test_id, default_rc=default_rc, default_stderr=default_stderr)
    mock_connection = MagicMock()
    mock_connection._shell.tmpdir = 'shell_tmpdir'
    mock_connection._shell.join_path = ntpath.join
    mock_connection.exec_command = progress_helper.poll

    return mock_connection


def win_updates_init(task_args, async_val=0, check_mode=False, connection=None):
    task = MagicMock(Task)
    task.args = task_args
    task.check_mode = check_mode
    task.async_val = async_val

    connection = connection or MagicMock()

    # Used for older Ansible versions
    play_context = MagicMock()
    play_context.check_mode = check_mode

    plugin = win_updates.ActionModule(task, connection, play_context, loader=None, templar=None,
                                      shared_loader_obj=None)
    return plugin


def run_action(monkeypatch, test_id, task_vars, check_mode=False, poll_rc=0, poll_stderr=b''):
    module_arg_return = task_vars.copy()
    module_arg_return['_wait'] = False

    plugin = win_updates_init(task_vars, check_mode=check_mode,
                              connection=mock_connection_init(test_id, default_rc=poll_rc, default_stderr=poll_stderr))
    execute_module = MagicMock()
    execute_module.return_value = {
        'invocation': {'module_args': module_arg_return},
        'output_path': 'update_output_path',
        'task_pid': 666,
        'cancel_id': 'update_cancel_id',
    }
    monkeypatch.setattr(plugin, '_execute_module', execute_module)
    monkeypatch.setattr(plugin, '_transfer_file', MagicMock())

    return plugin.run()


def test_run_with_async(monkeypatch):
    plugin = win_updates_init({}, async_val=1)
    execute_module = MagicMock()
    execute_module.return_value = {'invocation': {'module_args': {'_wait': True, 'accept_list': []}}, 'updates': ['test']}
    monkeypatch.setattr(plugin, '_execute_module', execute_module)

    # Running with async should just call the module and return back the result - sans the _wait invocation arg
    actual = plugin.run()
    assert actual == {'invocation': {'module_args': {'accept_list': []}}, 'updates': ['test']}

    assert execute_module.call_count == 1
    assert execute_module.call_args[1]['module_name'] == 'ansible.windows.win_updates'
    assert execute_module.call_args[1]['module_args']['_wait']
    assert execute_module.call_args[1]['module_args']['_output_path'] is None
    assert execute_module.call_args[1]['task_vars'] == {}


def test_failed_to_start_module(monkeypatch):
    plugin = win_updates_init({},)
    execute_module = MagicMock()
    execute_module.return_value = {
        'failed': True,
        'msg': 'Failed to start module details',
    }
    monkeypatch.setattr(plugin, '_execute_module', execute_module)
    monkeypatch.setattr(plugin, '_transfer_file', MagicMock())

    actual = plugin.run()
    assert actual['failed']
    assert actual['msg'] == 'Failed to start module details'
    assert 'exception' in actual
    assert actual['found_update_count'] == 0
    assert actual['failed_update_count'] == 0
    assert actual['installed_update_count'] == 0
    assert actual['filtered_updates'] == {}
    assert actual['updates'] == {}


def test_failure_with_poll_script(monkeypatch):
    actual = run_action(monkeypatch, 'fail_poll_script.txt', {}, poll_rc=1, poll_stderr=b'stderr msg')

    assert actual['rc'] == 1
    assert actual['stdout'] == 'stdout message\n14'
    assert actual['stderr'] == 'stderr msg'
    assert actual['failed']
    assert actual['msg'] == 'Failed to poll update task, see rc, stdout, stderr for more info'
    assert 'exception' in actual
    assert actual['found_update_count'] == 0
    assert actual['failed_update_count'] == 0
    assert actual['installed_update_count'] == 0
    assert actual['filtered_updates'] == {}
    assert actual['updates'] == {}


def test_poll_script_invalid_json_output(monkeypatch):
    actual = run_action(monkeypatch, 'fail_poll_script_invalid_json.txt', {}, poll_rc=0, poll_stderr=b'stderr msg')

    assert actual['rc'] == 0
    assert actual['stdout'] == '{"task":unquoted}\n17'
    assert actual['stderr'] == 'stderr msg'
    assert actual['failed']
    assert actual['msg'].startswith('Failed to decode poll result json: ')
    assert 'exception' in actual
    assert actual['found_update_count'] == 0
    assert actual['failed_update_count'] == 0
    assert actual['installed_update_count'] == 0
    assert actual['filtered_updates'] == {}
    assert actual['updates'] == {}


def test_install_with_multiple_reboots(monkeypatch):
    # Was tested against a Server 2019 host with the parameters set below.
    reboot_mock = MagicMock()
    reboot_mock.return_value = {'failed': False}
    monkeypatch.setattr(win_updates, 'reboot_host', reboot_mock)

    actual = run_action(monkeypatch, 'full_run.txt', {
        'category_names': '*',
        'state': 'installed',
        'reboot': 'yes',
    })

    assert reboot_mock.call_count == 2
    assert actual['changed']
    assert not actual['reboot_required']
    assert actual['found_update_count'] == 8
    assert actual['failed_update_count'] == 0
    assert actual['installed_update_count'] == 8
    assert actual['filtered_updates'] == {}
    assert len(actual['updates']) == 8

    for u_id, u in actual['updates'].items():
        assert u['id'] == u_id
        assert u['id'] in UPDATE_INFO
        u_info = UPDATE_INFO[u['id']]
        assert u['title'] == u_info['title']
        assert u['kb'] == [u_info['kb']]
        assert u['categories'] == u_info['categories']
        assert u['downloaded']
        assert u['installed']


def test_install_without_reboot(monkeypatch):
    reboot_mock = MagicMock()
    reboot_mock.return_value = {'failed': False}
    monkeypatch.setattr(win_updates, 'reboot_host', reboot_mock)

    actual = run_action(monkeypatch, 'install_no_reboot.txt', {
        'reject_list': ['KB2267602'],
    })

    assert reboot_mock.call_count == 0
    assert actual['changed']
    assert actual['reboot_required']
    assert actual['found_update_count'] == 3
    assert actual['failed_update_count'] == 0
    assert actual['installed_update_count'] == 3

    assert len(actual['filtered_updates']) == 3
    for u_id, u in actual['filtered_updates'].items():
        assert u['id'] == u_id
        assert u['id'] in UPDATE_INFO
        u_info = UPDATE_INFO[u['id']]
        assert u['title'] == u_info['title']
        assert u['kb'] == [u_info['kb']]
        assert u['categories'] == u_info['categories']
        assert not u['downloaded']
        assert not u['installed']

        if u_info['kb'] == '2267602':
            assert u['filtered_reason'] == 'blacklist'
            assert u['filtered_reasons'] == ['reject_list', 'category_names']
        else:
            assert u['filtered_reason'] == 'category_names'
            assert u['filtered_reasons'] == ['category_names']

    assert len(actual['updates']) == 3
    for u_id, u in actual['updates'].items():
        assert u['id'] == u_id
        assert u['id'] in UPDATE_INFO
        u_info = UPDATE_INFO[u['id']]
        assert u['title'] == u_info['title']
        assert u['kb'] == [u_info['kb']]
        assert u['categories'] == u_info['categories']
        assert u['downloaded']
        assert u['installed']


def test_install_with_initial_reboot_required(monkeypatch):
    reboot_mock = MagicMock()
    reboot_mock.return_value = {'failed': False}
    monkeypatch.setattr(win_updates, 'reboot_host', reboot_mock)

    actual = run_action(monkeypatch, 'install_reboot_req.txt', {
        'category_names': ['*'],
        'reboot': True,
    })

    assert reboot_mock.call_count == 2
    assert actual['changed']
    assert not actual['reboot_required']
    assert actual['found_update_count'] == 6
    assert actual['failed_update_count'] == 0
    assert actual['installed_update_count'] == 5  # 1 found update was already installed at the beginning
    assert actual['filtered_updates'] == {}

    assert len(actual['updates']) == 6
    for u_id, u in actual['updates'].items():
        assert u['id'] == u_id
        assert u['id'] in UPDATE_INFO
        u_info = UPDATE_INFO[u['id']]
        assert u['title'] == u_info['title']
        assert u['kb'] == [u_info['kb']]
        assert u['categories'] == u_info['categories']

        if u_info['kb'] == '5003171':
            assert not u['downloaded']
            assert not u['installed']
        else:
            assert u['downloaded']
            assert u['installed']


def test_install_with_reboot_fail(monkeypatch):
    reboot_mock = MagicMock()
    reboot_mock.return_value = {'failed': True, 'msg': 'Failure msg from reboot'}
    monkeypatch.setattr(win_updates, 'reboot_host', reboot_mock)

    actual = run_action(monkeypatch, 'install_reboot_req_failure.txt', {
        'category_names': ['*'],
        'reboot': True,
    })

    assert reboot_mock.call_count == 1
    assert not actual['changed']
    assert actual['reboot_required']
    assert actual['failed']
    assert actual['msg'] == 'Failed to reboot host: Failure msg from reboot'
    assert actual['found_update_count'] == 6
    assert actual['failed_update_count'] == 0
    assert actual['installed_update_count'] == 0
    assert actual['filtered_updates'] == {}

    assert len(actual['updates']) == 6
    for u_id, u in actual['updates'].items():
        assert u['id'] == u_id
        assert u['id'] in UPDATE_INFO
        u_info = UPDATE_INFO[u['id']]
        assert u['title'] == u_info['title']
        assert u['kb'] == [u_info['kb']]
        assert u['categories'] == u_info['categories']
        assert not u['downloaded']
        assert not u['installed']


def test_install_with_reboot_check_mode(monkeypatch):
    reboot_mock = MagicMock()
    reboot_mock.return_value = {'failed': True, 'msg': 'Failure msg from reboot'}
    monkeypatch.setattr(win_updates, 'reboot_host', reboot_mock)

    actual = run_action(monkeypatch, 'install_reboot_req_failure.txt', {
        'category_names': ['*'],
        'reboot': True,
    }, check_mode=True)

    assert reboot_mock.call_count == 0
    assert actual['changed']
    assert not actual['reboot_required']
    assert 'failed' not in actual
    assert 'msg' not in actual
    assert actual['found_update_count'] == 6
    assert actual['failed_update_count'] == 0
    assert actual['installed_update_count'] == 0
    assert actual['filtered_updates'] == {}

    assert len(actual['updates']) == 6
    for u_id, u in actual['updates'].items():
        assert u['id'] == u_id
        assert u['id'] in UPDATE_INFO
        u_info = UPDATE_INFO[u['id']]
        assert u['title'] == u_info['title']
        assert u['kb'] == [u_info['kb']]
        assert u['categories'] == u_info['categories']
        assert not u['downloaded']
        assert not u['installed']


def test_install_reboot_with_two_failures(monkeypatch):
    reboot_mock = MagicMock()
    reboot_mock.return_value = {'failed': False}
    monkeypatch.setattr(win_updates, 'reboot_host', reboot_mock)

    actual = run_action(monkeypatch, 'install_reboot_two_failures.txt', {
        'category_names': ['*'],
        'reboot': True,
    })

    assert reboot_mock.call_count == 1
    assert actual['changed']
    assert not actual['reboot_required']
    assert actual['failed']
    assert actual['msg'] == 'Searching for updates: Exception from HRESULT: 0x80240032 - The search criteria string was invalid ' \
        '(WU_E_INVALID_CRITERIA 80240032)'
    assert 'exception' in actual
    assert actual['found_update_count'] == 0
    assert actual['failed_update_count'] == 0
    assert actual['installed_update_count'] == 0
    assert actual['filtered_updates'] == {}
    assert actual['updates'] == {}


def test_install_with_initial_reboot_required_but_no_reboot(monkeypatch):
    reboot_mock = MagicMock()
    reboot_mock.return_value = {'failed': False}
    monkeypatch.setattr(win_updates, 'reboot_host', reboot_mock)

    actual = run_action(monkeypatch, 'install_reboot_req_no_reboot.txt', {
        'category_names': ['*'],
    })

    assert reboot_mock.call_count == 0
    assert not actual['changed']
    assert actual['failed']
    assert actual['msg'] == 'A reboot is required before more updates can be installed'
    assert actual['reboot_required']
    assert actual['found_update_count'] == 5
    assert actual['failed_update_count'] == 0
    assert actual['installed_update_count'] == 0
    assert actual['filtered_updates'] == {}

    assert len(actual['updates']) == 5
    for u_id, u in actual['updates'].items():
        assert u['id'] == u_id
        assert u['id'] in UPDATE_INFO
        u_info = UPDATE_INFO[u['id']]
        assert u['title'] == u_info['title']
        assert u['kb'] == [u_info['kb']]
        assert u['categories'] == u_info['categories']
        assert not u['downloaded']
        assert not u['installed']


def test_install_non_integer_kb_values(monkeypatch):
    # custom_kb.txt contains 3 updates and 3 filtered updates an empty kb, integer kb, and string kb value.
    actual = run_action(monkeypatch, 'custom_kb.txt', {
        'reject_list': ['KB2267602'],
    })

    assert len(actual['filtered_updates']) == 3
    for u_id, u in actual['filtered_updates'].items():
        assert u['id'] == u_id
        if u['id'] == 'f26a0046-1e1a-4305-8743-19c92c3095a5':
            assert u['kb'] == ['']

        elif u['id'] == '9e94ae7a-f0f4-4ea1-b590-9f1388ee6126':
            assert u['kb'] == ['5003396']

        elif u['id'] == '4326d7df-c256-4cca-99f7-c02a04443ec1':
            assert u['kb'] == ['OTHER']

        else:
            assert False, ("Unknown update in filtered updates %s" % u['id'])

    assert len(actual['updates']) == 3
    for u_id, u in actual['updates'].items():
        assert u['id'] == u_id
        if u['id'] == '74819184-828b-4f97-bb4b-089a0c58e366':
            assert u['kb'] == ['']

        elif u['id'] == 'd4919b6d-584d-436f-b877-dc3fc352a774':
            assert u['kb'] == ['890830']

        elif u['id'] == '1abb2377-20ef-43ff-aabc-0de4711ab205':
            assert u['kb'] == ['NET4800']

        else:
            assert False, ("Unknown update in filtered updates %s" % u['id'])


def test_fail_install(monkeypatch):
    reboot_mock = MagicMock()
    reboot_mock.return_value = {'failed': False}
    monkeypatch.setattr(win_updates, 'reboot_host', reboot_mock)

    actual = run_action(monkeypatch, 'failed_install.txt', {
        'category_names': ['SecurityUpdates', 'DefinitionUpdates'],
        'accept_list': ['KB5003171', 'KB2267602'],
    })

    assert reboot_mock.call_count == 0
    assert actual['changed']
    assert actual['failed']
    assert actual['reboot_required']
    assert actual['found_update_count'] == 2
    assert actual['failed_update_count'] == 1
    assert actual['installed_update_count'] == 1

    assert len(actual['filtered_updates']) == 4
    for u_id, u in actual['filtered_updates'].items():
        assert u['id'] == u_id
        assert u['id'] in UPDATE_INFO
        u_info = UPDATE_INFO[u['id']]
        assert u['title'] == u_info['title']
        assert u['kb'] == [u_info['kb']]
        assert u['categories'] == u_info['categories']
        assert not u['downloaded']
        assert not u['installed']

        assert u['filtered_reason'] == 'whitelist'
        if u_info['kb'] == '4580325':
            assert u['filtered_reasons'] == ['accept_list']
        else:
            assert u['filtered_reasons'] == ['accept_list', 'category_names']

    assert len(actual['updates']) == 2
    for u_id, u in actual['updates'].items():
        assert u['id'] == u_id
        assert u['id'] in UPDATE_INFO
        u_info = UPDATE_INFO[u['id']]
        assert u['title'] == u_info['title']
        assert u['kb'] == [u_info['kb']]
        assert u['categories'] == u_info['categories']
        assert u['downloaded']

        if u_info['kb'] == '2267602':
            assert not u['installed']
            assert u['failure_hresult_code'] == 2147944003
            assert u['failure_msg'] == 'Unknown WUA HRESULT 2147944003 (UNKNOWN 80070643)'

        else:
            assert u['installed']
            assert 'failure_hresult_code' not in u



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: projects_details_info
short_description: Information module for Projects Details
description:
- Get all Projects Details.
- Get projects details.
version_added: '4.0.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  id:
    description:
    - Id query parameter. Id of project to be searched.
    type: str
  name:
    description:
    - Name query parameter. Name of project to be searched.
    type: str
  offset:
    description:
    - Offset query parameter. Index of first result.
    type: int
  limit:
    description:
    - Limit query parameter. Limits number of results.
    type: int
  sortOrder:
    description:
    - SortOrder query parameter. Sort Order Ascending (asc) or Descending (dsc).
    type: str
requirements:
- dnacentersdk >= 2.5.0
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Configuration Templates GetProjectsDetails
  description: Complete reference of the GetProjectsDetails API.
  link: https://developer.cisco.com/docs/dna-center/#!get-project-s-details
notes:
  - SDK Method used are
    configuration_templates.ConfigurationTemplates.get_projects_details,

  - Paths used are
    get /dna/intent/api/v2/template-programmer/project,

"""

EXAMPLES = r"""
- name: Get all Projects Details
  cisco.dnac.projects_details_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    id: string
    name: string
    offset: 0
    limit: 0
    sortOrder: string
  register: result

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": [
        {
          "createTime": 0,
          "description": "string",
          "id": "string",
          "isDeletable": true,
          "lastUpdateTime": 0,
          "name": "string",
          "tags": [
            {
              "id": "string",
              "name": "string"
            }
          ],
          "templates": [
            {
              "tags": [
                {
                  "id": "string",
                  "name": "string"
                }
              ],
              "author": "string",
              "composite": true,
              "containingTemplates": [
                {
                  "tags": [
                    {
                      "id": "string",
                      "name": "string"
                    }
                  ],
                  "composite": true,
                  "description": "string",
                  "deviceTypes": [
                    {
                      "productFamily": "string",
                      "productSeries": "string",
                      "productType": "string"
                    }
                  ],
                  "id": "string",
                  "language": "string",
                  "name": "string",
                  "projectName": "string",
                  "rollbackTemplateParams": [
                    {
                      "binding": "string",
                      "customOrder": 0,
                      "dataType": "string",
                      "defaultValue": "string",
                      "description": "string",
                      "displayName": "string",
                      "group": "string",
                      "id": "string",
                      "instructionText": "string",
                      "key": "string",
                      "notParam": true,
                      "order": 0,
                      "paramArray": true,
                      "parameterName": "string",
                      "provider": "string",
                      "range": [
                        {
                          "id": "string",
                          "maxValue": 0,
                          "minValue": 0
                        }
                      ],
                      "required": true,
                      "selection": {
                        "defaultSelectedValues": [
                          "string"
                        ],
                        "id": "string",
                        "selectionType": "string",
                        "selectionValues": {}
                      }
                    }
                  ],
                  "templateContent": "string",
                  "templateParams": [
                    {
                      "binding": "string",
                      "customOrder": 0,
                      "dataType": "string",
                      "defaultValue": "string",
                      "description": "string",
                      "displayName": "string",
                      "group": "string",
                      "id": "string",
                      "instructionText": "string",
                      "key": "string",
                      "notParam": true,
                      "order": 0,
                      "paramArray": true,
                      "parameterName": "string",
                      "provider": "string",
                      "range": [
                        {
                          "id": "string",
                          "maxValue": 0,
                          "minValue": 0
                        }
                      ],
                      "required": true,
                      "selection": {
                        "defaultSelectedValues": [
                          "string"
                        ],
                        "id": "string",
                        "selectionType": "string",
                        "selectionValues": {}
                      }
                    }
                  ],
                  "version": "string"
                }
              ],
              "createTime": 0,
              "customParamsOrder": true,
              "description": "string",
              "deviceTypes": [
                {
                  "productFamily": "string",
                  "productSeries": "string",
                  "productType": "string"
                }
              ],
              "failurePolicy": "string",
              "id": "string",
              "language": "string",
              "lastUpdateTime": 0,
              "latestVersionTime": 0,
              "name": "string",
              "parentTemplateId": "string",
              "projectId": "string",
              "projectName": "string",
              "rollbackTemplateContent": "string",
              "rollbackTemplateParams": [
                {
                  "binding": "string",
                  "customOrder": 0,
                  "dataType": "string",
                  "defaultValue": "string",
                  "description": "string",
                  "displayName": "string",
                  "group": "string",
                  "id": "string",
                  "instructionText": "string",
                  "key": "string",
                  "notParam": true,
                  "order": 0,
                  "paramArray": true,
                  "parameterName": "string",
                  "provider": "string",
                  "range": [
                    {
                      "id": "string",
                      "maxValue": 0,
                      "minValue": 0
                    }
                  ],
                  "required": true,
                  "selection": {
                    "defaultSelectedValues": [
                      "string"
                    ],
                    "id": "string",
                    "selectionType": "string",
                    "selectionValues": {}
                  }
                }
              ],
              "softwareType": "string",
              "softwareVariant": "string",
              "softwareVersion": "string",
              "templateContent": "string",
              "templateParams": [
                {
                  "binding": "string",
                  "customOrder": 0,
                  "dataType": "string",
                  "defaultValue": "string",
                  "description": "string",
                  "displayName": "string",
                  "group": "string",
                  "id": "string",
                  "instructionText": "string",
                  "key": "string",
                  "notParam": true,
                  "order": 0,
                  "paramArray": true,
                  "parameterName": "string",
                  "provider": "string",
                  "range": [
                    {
                      "id": "string",
                      "maxValue": 0,
                      "minValue": 0
                    }
                  ],
                  "required": true,
                  "selection": {
                    "defaultSelectedValues": [
                      "string"
                    ],
                    "id": "string",
                    "selectionType": "string",
                    "selectionValues": {}
                  }
                }
              ],
              "validationErrors": {
                "rollbackTemplateErrors": [
                  {}
                ],
                "templateErrors": [
                  {}
                ],
                "templateId": "string",
                "templateVersion": "string"
              },
              "version": "string"
            }
          ]
        }
      ],
      "version": "string"
    }
"""



#
# (c) 2016 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, division, print_function


__metaclass__ = type

import copy
import re
import sys

from ansible import constants as C
from ansible.module_utils.connection import Connection
from ansible.utils.display import Display
from ansible_collections.ansible.netcommon.plugins.action.network import (
    ActionModule as ActionNetworkModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    load_provider,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.nxos import nxos_provider_spec


display = Display()


class ActionModule(ActionNetworkModule):
    def run(self, tmp=None, task_vars=None):
        del tmp  # tmp no longer has any effect

        module_name = self._task.action.split(".")[-1]
        self._config_module = True if module_name in ["nxos_config", "config"] else False
        persistent_connection = self._play_context.connection.split(".")[-1]

        warnings = []

        if (
            self._play_context.connection in ("httpapi", "local")
            or self._task.args.get("provider", {}).get("transport") == "nxapi"
        ) and module_name in ("nxos_file_copy", "nxos_nxapi"):
            return {
                "failed": True,
                "msg": "Transport type 'nxapi' is not valid for '%s' module." % (module_name),
            }

        if module_name == "nxos_file_copy":
            # when file_pull is enabled, the file_pull_timeout and connect_ssh_port options
            # will override persistent_command_timeout and port
            # this has been kept for backwards compatibility till these options are removed
            if persistent_connection != "network_cli":
                return {
                    "failed": True,
                    "msg": "Connection type must be fully qualified name for network_cli connection type, got %s"
                    % self._play_context.connection,
                }

            conn = Connection(self._connection.socket_path)
            file_pull = self._task.args.get("file_pull", False)
            file_pull_timeout = self._task.args.get("file_pull_timeout", 300)
            connect_ssh_port = self._task.args.get("connect_ssh_port", 22)

            if file_pull:
                conn.set_option("persistent_command_timeout", file_pull_timeout)
                conn.set_option("port", connect_ssh_port)

        if module_name == "nxos_install_os":
            connection = self._connection
            if connection.transport == "local":
                persistent_command_timeout = C.PERSISTENT_COMMAND_TIMEOUT
                persistent_connect_timeout = C.PERSISTENT_CONNECT_TIMEOUT
            else:
                persistent_command_timeout = connection.get_option("persistent_command_timeout")
                persistent_connect_timeout = connection.get_option("persistent_connect_timeout")

            display.vvvv(
                "PERSISTENT_COMMAND_TIMEOUT is %s" % str(persistent_command_timeout),
                self._play_context.remote_addr,
            )
            display.vvvv(
                "PERSISTENT_CONNECT_TIMEOUT is %s" % str(persistent_connect_timeout),
                self._play_context.remote_addr,
            )
            if persistent_command_timeout < 600 or persistent_connect_timeout < 600:
                msg = "PERSISTENT_COMMAND_TIMEOUT and PERSISTENT_CONNECT_TIMEOUT"
                msg += " must be set to 600 seconds or higher when using nxos_install_os module."
                msg += " Current persistent_command_timeout setting:" + str(
                    persistent_command_timeout,
                )
                msg += " Current persistent_connect_timeout setting:" + str(
                    persistent_connect_timeout,
                )
                return {"failed": True, "msg": msg}

        if persistent_connection in ("network_cli", "httpapi"):
            provider = self._task.args.get("provider", {})
            if any(provider.values()):
                display.warning(
                    "provider is unnecessary when using %s and will be ignored"
                    % self._play_context.connection,
                )
                del self._task.args["provider"]

            if module_name == "nxos_gir":
                conn = Connection(self._connection.socket_path)
                persistent_command_timeout = conn.get_option("persistent_command_timeout")
                gir_timeout = 200
                if persistent_command_timeout < gir_timeout:
                    conn.set_option("persistent_command_timeout", gir_timeout)
                    msg = "timeout value extended to %ss for nxos_gir" % gir_timeout
                    display.warning(msg)

        elif self._play_context.connection == "local":
            provider = load_provider(nxos_provider_spec, self._task.args)
            transport = provider["transport"] or "cli"

            display.vvvv(
                "connection transport is %s" % transport,
                self._play_context.remote_addr,
            )

            if transport == "cli":
                pc = copy.deepcopy(self._play_context)
                pc.connection = "ansible.netcommon.network_cli"
                pc.network_os = "cisco.nxos.nxos"
                pc.remote_addr = provider["host"] or self._play_context.remote_addr
                pc.port = int(provider["port"] or self._play_context.port or 22)
                pc.remote_user = provider["username"] or self._play_context.connection_user
                pc.password = provider["password"] or self._play_context.password
                pc.private_key_file = provider["ssh_keyfile"] or self._play_context.private_key_file
                pc.become = provider["authorize"] or False
                if pc.become:
                    pc.become_method = "enable"
                pc.become_pass = provider["auth_pass"]

                connection = self._shared_loader_obj.connection_loader.get(
                    "ansible.netcommon.persistent",
                    pc,
                    sys.stdin,
                    task_uuid=self._task._uuid,
                )

                # TODO: Remove below code after ansible minimal is cut out
                if connection is None:
                    pc.connection = "network_cli"
                    pc.network_os = "nxos"
                    connection = self._shared_loader_obj.connection_loader.get(
                        "persistent",
                        pc,
                        sys.stdin,
                        task_uuid=self._task._uuid,
                    )

                display.vvv(
                    "using connection plugin %s (was local)" % pc.connection,
                    pc.remote_addr,
                )

                command_timeout = (
                    int(provider["timeout"])
                    if provider["timeout"]
                    else connection.get_option("persistent_command_timeout")
                )
                connection.set_options(direct={"persistent_command_timeout": command_timeout})

                socket_path = connection.run()
                display.vvvv("socket_path: %s" % socket_path, pc.remote_addr)
                if not socket_path:
                    return {
                        "failed": True,
                        "msg": "unable to open shell. Please see: "
                        + "https://docs.ansible.com/ansible/network_debug_troubleshooting.html#unable-to-open-shell",
                    }

                task_vars["ansible_socket"] = socket_path

            else:
                self._task.args["provider"] = ActionModule.nxapi_implementation(
                    provider,
                    self._play_context,
                )
                warnings.append(
                    [
                        "connection local support for this module is deprecated and will be removed in version 2.14,"
                        " use connection either httpapi or ansible.netcommon.httpapi (whichever is applicable)",
                    ],
                )
        else:
            return {
                "failed": True,
                "msg": "Connection type %s is not valid for this module"
                % self._play_context.connection,
            }

        result = super(ActionModule, self).run(task_vars=task_vars)
        if warnings:
            if "warnings" in result:
                result["warnings"].extend(warnings)
            else:
                result["warnings"] = warnings
        return result

    @staticmethod
    def nxapi_implementation(provider, play_context):
        provider["transport"] = "nxapi"
        if provider.get("host") is None:
            provider["host"] = play_context.remote_addr

        if provider.get("port") is None:
            if provider.get("use_ssl"):
                provider["port"] = 443
            else:
                provider["port"] = 80

        if provider.get("timeout") is None:
            provider["timeout"] = C.PERSISTENT_COMMAND_TIMEOUT

        if provider.get("username") is None:
            provider["username"] = play_context.connection_user

        if provider.get("password") is None:
            provider["password"] = play_context.password

        if provider.get("use_ssl") is None:
            provider["use_ssl"] = False

        if provider.get("validate_certs") is None:
            provider["validate_certs"] = True

        return provider



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: sda_fabric_info
short_description: Information module for Sda Fabric
description:
- Get all Sda Fabric.
- Get SDA Fabric Info.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  fabricName:
    description:
    - FabricName query parameter. Fabric Name.
    type: str
requirements:
- dnacentersdk >= 2.4.9
- python >= 3.5
notes:
  - SDK Method used are
    sda.Sda.get_sda_fabric_info,

  - Paths used are
    get /dna/intent/api/v1/business/sda/fabric,

"""

EXAMPLES = r"""
- name: Get all Sda Fabric
  cisco.dnac.sda_fabric_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers:
      custom: value
    fabricName: string
  register: result

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "status": "string",
      "description": "string",
      "executionId": "string",
      "fabricName": "string",
      "fabricType": "string",
      "fabricDomainType": "string"
    }
"""



#!/usr/bin/python
# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = r'''
---
module: elb_instance
version_added: 1.0.0
short_description: De-registers or registers instances from EC2 ELBs
description:
  - This module de-registers or registers an AWS EC2 instance from the ELBs
    that it belongs to.
  - Will be marked changed when called only if there are ELBs found to operate on.
author: "John Jarvis (@jarv)"
options:
  state:
    description:
      - Register or deregister the instance.
    required: true
    choices: ['present', 'absent']
    type: str
  instance_id:
    description:
      - EC2 Instance ID.
    required: true
    type: str
  ec2_elbs:
    description:
      - List of ELB names
      - Required when I(state=present).
    type: list
    elements: str
  enable_availability_zone:
    description:
      - Whether to enable the availability zone of the instance on the target ELB if the availability zone has not already
        been enabled.
      - If I(enable_availability_zone=no), the task will fail if the availability zone is not enabled on the ELB.
    type: bool
    default: 'yes'
  wait:
    description:
      - Wait for instance registration or deregistration to complete successfully before returning.
    type: bool
    default: 'yes'
  wait_timeout:
    description:
      - Number of seconds to wait for an instance to change state.
      - If I(wait_timeout=0) then this module may return an error if a transient error occurs.
      - If non-zero then any transient errors are ignored until the timeout is reached.
      - Ignored when I(wait=no).
    default: 0
    type: int
notes:
- The ec2_elb fact currently set by this module has been deprecated and will no
  longer be set after release 4.0.0 of the collection.
extends_documentation_fragment:
- amazon.aws.aws
- amazon.aws.ec2
'''

EXAMPLES = r"""
# basic pre_task and post_task example
pre_tasks:
  - name: Instance De-register
    community.aws.elb_instance:
      instance_id: "{{ ansible_ec2_instance_id }}"
      state: absent
    register: deregister_instances
    delegate_to: localhost
roles:
  - myrole
post_tasks:
  - name: Instance Register
    community.aws.elb_instance:
      instance_id: "{{ ansible_ec2_instance_id }}"
      ec2_elbs: "{{ deregister_instances.updated_elbs }}"
      state: present
    delegate_to: localhost
"""

RETURN = '''
updated_elbs:
  description: A list of ELB names that the instance has been added to or removed from.
  returned: always
  type: list
  elements: str
'''

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible_collections.amazon.aws.plugins.module_utils.core import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.core import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AWSRetry


class ElbManager:
    """Handles EC2 instance ELB registration and de-registration"""

    def __init__(self, module, instance_id=None, ec2_elbs=None):
        retry_decorator = AWSRetry.jittered_backoff()
        self.module = module
        self.client_asg = module.client('autoscaling', retry_decorator=retry_decorator)
        self.client_ec2 = module.client('ec2', retry_decorator=retry_decorator)
        self.client_elb = module.client('elb', retry_decorator=retry_decorator)
        self.instance_id = instance_id
        self.lbs = self._get_instance_lbs(ec2_elbs)
        self.changed = False
        self.updated_elbs = set()

    def deregister(self, wait, timeout):
        """De-register the instance from all ELBs and wait for the ELB
        to report it out-of-service"""

        for lb in self.lbs:
            instance_ids = [i['InstanceId'] for i in lb['Instances']]
            if self.instance_id not in instance_ids:
                continue

            self.updated_elbs.add(lb['LoadBalancerName'])

            if self.module.check_mode:
                self.changed = True
                continue

            try:
                self.client_elb.deregister_instances_from_load_balancer(
                    aws_retry=True,
                    LoadBalancerName=lb['LoadBalancerName'],
                    Instances=[{"InstanceId": self.instance_id}],
                )
            except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
                self.module.fail_json_aws(e, 'Failed to deregister instance from load balancer',
                                          load_balancer=lb, instance=self.instance_id)

            # The ELB is changing state in some way. Either an instance that's
            # InService is moving to OutOfService, or an instance that's
            # already OutOfService is being deregistered.
            self.changed = True

        if wait:
            for lb in self.lbs:
                self._await_elb_instance_state(lb, 'Deregistered', timeout)

    def register(self, wait, enable_availability_zone, timeout):
        """Register the instance for all ELBs and wait for the ELB
        to report the instance in-service"""
        for lb in self.lbs:
            instance_ids = [i['InstanceId'] for i in lb['Instances']]
            if self.instance_id in instance_ids:
                continue

            self.updated_elbs.add(lb['LoadBalancerName'])

            if enable_availability_zone:
                self.changed |= self._enable_availailability_zone(lb)

            if self.module.check_mode:
                self.changed = True
                continue

            try:
                self.client_elb.register_instances_with_load_balancer(
                    aws_retry=True,
                    LoadBalancerName=lb['LoadBalancerName'],
                    Instances=[{"InstanceId": self.instance_id}],
                )
            except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
                self.module.fail_json_aws(e, 'Failed to register instance with load balancer',
                                          load_balancer=lb, instance=self.instance_id)

            self.changed = True

        if wait:
            for lb in self.lbs:
                self._await_elb_instance_state(lb, 'InService', timeout)

    @AWSRetry.jittered_backoff()
    def _describe_elbs(self, **params):
        paginator = self.client_elb.get_paginator('describe_load_balancers')
        results = paginator.paginate(**params).build_full_result()
        return results['LoadBalancerDescriptions']

    def exists(self, lbtest):
        """ Verify that the named ELB actually exists """

        found = False
        for lb in self.lbs:
            if lb['LoadBalancerName'] == lbtest:
                found = True
                break
        return found

    def _enable_availailability_zone(self, lb):
        """Enable the current instance's availability zone in the provided lb.
        Returns True if the zone was enabled or False if no change was made.
        lb: load balancer"""
        instance = self._get_instance()
        desired_zone = instance['Placement']['AvailabilityZone']

        if desired_zone in lb['AvailabilityZones']:
            return False

        if self.module.check_mode:
            return True

        try:
            self.client_elb.enable_availability_zones_for_load_balancer(
                aws_retry=True,
                LoadBalancerName=lb['LoadBalancerName'],
                AvailabilityZones=[desired_zone],
            )
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            self.module.fail_json_aws(e, 'Failed to enable AZ on load balancers',
                                      load_balancer=lb, zone=desired_zone)

        return True

    def _await_elb_instance_state(self, lb, awaited_state, timeout):
        """Wait for an ELB to change state"""
        if self.module.check_mode:
            return

        initial_state = self._get_instance_health(lb)

        if awaited_state == initial_state:
            return

        if awaited_state == 'InService':
            waiter = self.client_elb.get_waiter('instance_in_service')
        elif awaited_state == 'Deregistered':
            waiter = self.client_elb.get_waiter('instance_deregistered')
        elif awaited_state == 'OutOfService':
            waiter = self.client_elb.get_waiter('instance_deregistered')
        else:
            self.module.fail_json(msg='Could not wait for unknown state', awaited_state=awaited_state)

        try:
            waiter.wait(
                LoadBalancerName=lb['LoadBalancerName'],
                Instances=[{"InstanceId": self.instance_id}],
                WaiterConfig={'Delay': 1, 'MaxAttempts': timeout},
            )
        except botocore.exceptions.WaiterError as e:
            self.module.fail_json_aws(e, msg='Timeout waiting for instance to reach desired state',
                                      awaited_state=awaited_state)
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            self.module.fail_json_aws(e, msg='Error while waiting for instance to reach desired state',
                                      awaited_state=awaited_state)

        return

    def _get_instance_health(self, lb):
        """
        Check instance health, should return status object or None under
        certain error conditions.
        """
        try:
            status = self.client_elb.describe_instance_health(
                aws_retry=True,
                LoadBalancerName=lb['LoadBalancerName'],
                Instances=[{'InstanceId': self.instance_id}],
            )['InstanceStates']
        except is_boto3_error_code('InvalidInstance'):
            return None
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:  # pylint: disable=duplicate-except
            self.module.fail_json_aws(e, msg='Failed to get instance health')

        if not status:
            return None

        return status[0]['State']

    def _get_instance_lbs(self, ec2_elbs=None):
        """Returns a list of ELBs attached to self.instance_id
        ec2_elbs: an optional list of elb names that will be used
                  for elb lookup instead of returning what elbs
                  are attached to self.instance_id"""

        list_params = dict()
        if not ec2_elbs:
            ec2_elbs = self._get_auto_scaling_group_lbs()

        if ec2_elbs:
            list_params['LoadBalancerNames'] = ec2_elbs

        try:
            elbs = self._describe_elbs(**list_params)
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            self.module.fail_json_aws(e, 'Failed to describe load balancers')

        if ec2_elbs:
            return elbs

        # If ec2_elbs wasn't specified, then filter out LBs we're not a member
        # of.
        lbs = []
        for lb in elbs:
            instance_ids = [i['InstanceId'] for i in lb['Instances']]
            if self.instance_id in instance_ids:
                lbs.append(lb)

        return lbs

    def _get_auto_scaling_group_lbs(self):
        """Returns a list of ELBs associated with self.instance_id
           indirectly through its auto scaling group membership"""

        try:
            asg_instances = self.client_asg.describe_auto_scaling_instances(
                aws_retry=True,
                InstanceIds=[self.instance_id])['AutoScalingInstances']
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            self.module.fail_json_aws(e, msg='Failed to describe ASG Instance')

        if len(asg_instances) > 1:
            self.module.fail_json(msg="Illegal state, expected one auto scaling group instance.")

        if not asg_instances:
            # Instance isn't a member of an ASG
            return []

        asg_name = asg_instances[0]['AutoScalingGroupName']

        try:
            asg_instances = self.client_asg.describe_auto_scaling_groups(
                aws_retry=True,
                AutoScalingGroupNames=[asg_name])['AutoScalingGroups']
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            self.module.fail_json_aws(e, msg='Failed to describe ASG Instance')

        if len(asg_instances) != 1:
            self.module.fail_json(msg="Illegal state, expected one auto scaling group.")

        return asg_instances[0]['LoadBalancerNames']

    def _get_instance(self):
        """Returns the description of an instance"""
        try:
            result = self.client_ec2.describe_instances(
                aws_retry=True,
                InstanceIds=[self.instance_id])
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            self.module.fail_json_aws(e, msg='Failed to describe ASG Instance')
        return result['Reservations'][0]['Instances'][0]


def main():
    argument_spec = dict(
        state={'required': True, 'choices': ['present', 'absent']},
        instance_id={'required': True},
        ec2_elbs={'default': None, 'required': False, 'type': 'list', 'elements': 'str'},
        enable_availability_zone={'default': True, 'required': False, 'type': 'bool'},
        wait={'required': False, 'default': True, 'type': 'bool'},
        wait_timeout={'required': False, 'default': 0, 'type': 'int'},
    )
    required_if = [
        ('state', 'present', ['ec2_elbs']),
    ]

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        required_if=required_if,
        supports_check_mode=True,
    )

    ec2_elbs = module.params['ec2_elbs']
    wait = module.params['wait']
    enable_availability_zone = module.params['enable_availability_zone']
    timeout = module.params['wait_timeout']
    instance_id = module.params['instance_id']

    elb_man = ElbManager(module, instance_id, ec2_elbs)

    if ec2_elbs is not None:
        for elb in ec2_elbs:
            if not elb_man.exists(elb):
                module.fail_json(msg="ELB {0} does not exist".format(elb))

    if module.params['state'] == 'present':
        elb_man.register(wait, enable_availability_zone, timeout)
    elif module.params['state'] == 'absent':
        elb_man.deregister(wait, timeout)

    # XXX We're not an _fact module we shouldn't be returning a fact and poluting
    # the namespace
    ansible_facts = {'ec2_elbs': [lb['LoadBalancerName'] for lb in elb_man.lbs]}

    module.exit_json(
        changed=elb_man.changed,
        ansible_facts=ansible_facts,
        updated_elbs=list(elb_man.updated_elbs),
    )


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# (c) 2020, Jordan Borean <jborean93@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

import re
import sys

import cryptography
import pytest

from cryptography.x509 import NameAttribute, oid

from ansible_collections.community.crypto.plugins.module_utils.crypto.basic import (
    OpenSSLObjectError,
)

from ansible_collections.community.crypto.plugins.module_utils.crypto.cryptography_support import (
    cryptography_get_name,
    _adjust_idn,
    _parse_dn_component,
    _parse_dn,
)

from ansible_collections.community.crypto.plugins.module_utils.version import LooseVersion


@pytest.mark.parametrize('unicode, idna, cycled_unicode', [
    (u'..', u'..', None),
    (u'foo.com', u'foo.com', None),
    (u'.foo.com.', u'.foo.com.', None),
    (u'*.foo.com', u'*.foo.com', None),
    (u'straße', u'xn--strae-oqa', None),
    (u'ﬀóò.ḃâŗ.çøṁ', u'xn--ff-3jad.xn--2ca8uh37e.xn--7ca8a981n', u'ffóò.ḃâŗ.çøṁ'),
    (u'*.☺.', u'*.xn--74h.', None),
])
def test_adjust_idn(unicode, idna, cycled_unicode):
    if cycled_unicode is None:
        cycled_unicode = unicode

    result = _adjust_idn(unicode, 'ignore')
    print(result, unicode)
    assert result == unicode

    result = _adjust_idn(idna, 'ignore')
    print(result, idna)
    assert result == idna

    result = _adjust_idn(unicode, 'unicode')
    print(result, unicode)
    assert result == unicode

    result = _adjust_idn(idna, 'unicode')
    print(result, cycled_unicode)
    assert result == cycled_unicode

    result = _adjust_idn(unicode, 'idna')
    print(result, idna)
    assert result == idna

    result = _adjust_idn(idna, 'idna')
    print(result, idna)
    assert result == idna


@pytest.mark.parametrize('value, idn_rewrite, message', [
    (u'bar', 'foo', re.escape(u'Invalid value for idn_rewrite: "foo"')),
])
def test_adjust_idn_fail_valueerror(value, idn_rewrite, message):
    with pytest.raises(ValueError, match=message):
        result = _adjust_idn(value, idn_rewrite)


@pytest.mark.parametrize('value, idn_rewrite, message', [
    (
        u'xn--a',
        'unicode',
        u'''^Error while transforming part u?"xn\\-\\-a" of IDNA DNS name u?"xn\\-\\-a" to Unicode\\.'''
        u''' IDNA2008 transformation resulted in "Codepoint U\\+0080 at position 1 of u?'\\\\x80' not allowed",'''
        u''' IDNA2003 transformation resulted in "(decoding with 'idna' codec failed'''
        u''' \\(UnicodeError: )?Invalid character u?'\\\\x80'\\)?"\\.$'''
    ),
])
def test_adjust_idn_fail_user_error(value, idn_rewrite, message):
    with pytest.raises(OpenSSLObjectError, match=message):
        result = _adjust_idn(value, idn_rewrite)


def test_cryptography_get_name_invalid_prefix():
    with pytest.raises(OpenSSLObjectError, match="^Cannot parse Subject Alternative Name"):
        cryptography_get_name('fake:value')


def test_cryptography_get_name_other_name_no_oid():
    with pytest.raises(OpenSSLObjectError, match="Cannot parse Subject Alternative Name otherName"):
        cryptography_get_name('otherName:value')


def test_cryptography_get_name_other_name_utfstring():
    actual = cryptography_get_name('otherName:1.3.6.1.4.1.311.20.2.3;UTF8:Hello World')
    assert actual.type_id.dotted_string == '1.3.6.1.4.1.311.20.2.3'
    assert actual.value == b'\x0c\x0bHello World'


@pytest.mark.parametrize('name, options, expected', [
    (b'CN=x ', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u'x '), b'')),
    (b'CN=\\ ', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u' '), b'')),
    (b'CN=\\#', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u'#'), b'')),
    (b'CN=#402032', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u'@ 2'), b'')),
    (b'CN = x ', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u'x '), b'')),
    (b'CN = x\\, ', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u'x, '), b'')),
    (b'CN = x\\40 ', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u'x@ '), b'')),
    (b'CN  =  \\  , / ', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u'  '), b', / ')),
    (b'CN  =  \\  , / ', {'sep': b'/'}, (NameAttribute(oid.NameOID.COMMON_NAME, u'  , '), b'/ ')),
    (b'CN  =  \\  , / ', {'decode_remainder': False}, (NameAttribute(oid.NameOID.COMMON_NAME, u'\\  , / '), b'')),
    # Some examples from https://datatracker.ietf.org/doc/html/rfc4514#section-4:
    (b'CN=James \\"Jim\\" Smith\\, III', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u'James "Jim" Smith, III'), b'')),
    (b'CN=Before\\0dAfter', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u'Before\x0dAfter'), b'')),
    (b'1.3.6.1.4.1.1466.0=#04024869', {}, (NameAttribute(oid.ObjectIdentifier(u'1.3.6.1.4.1.1466.0'), u'\x04\x02Hi'), b'')),
    (b'CN=Lu\\C4\\8Di\\C4\\87', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u'Lučić'), b'')),
])
def test_parse_dn_component(name, options, expected):
    result = _parse_dn_component(name, **options)
    print(result, expected)
    assert result == expected


# Cryptography < 2.9 does not allow empty strings
# (https://github.com/pyca/cryptography/commit/87b2749c52e688c809f1861e55d958c64147493c)
if LooseVersion(cryptography.__version__) >= LooseVersion('2.9'):
    @pytest.mark.parametrize('name, options, expected', [
        (b'CN=', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u''), b'')),
        (b'CN= ', {}, (NameAttribute(oid.NameOID.COMMON_NAME, u''), b'')),
    ])
    def test_parse_dn_component_not_py26(name, options, expected):
        result = _parse_dn_component(name, **options)
        print(result, expected)
        assert result == expected


@pytest.mark.parametrize('name, options, message', [
    (b'CN=\\0', {}, u'Hex escape sequence "\\0" incomplete at end of string'),
    (b'CN=\\0,', {}, u'Hex escape sequence "\\0," has invalid second letter'),
    (b'CN=#0,', {}, u'Invalid hex sequence entry "0,"'),
])
def test_parse_dn_component_failure(name, options, message):
    with pytest.raises(OpenSSLObjectError, match=u'^%s$' % re.escape(message)):
        result = _parse_dn_component(name, **options)


@pytest.mark.parametrize('name, expected', [
    (b'CN=foo', [NameAttribute(oid.NameOID.COMMON_NAME, u'foo')]),
    (b'CN=foo,CN=bar', [NameAttribute(oid.NameOID.COMMON_NAME, u'foo'), NameAttribute(oid.NameOID.COMMON_NAME, u'bar')]),
    (b'CN  =  foo ,  CN  =  bar', [NameAttribute(oid.NameOID.COMMON_NAME, u'foo '), NameAttribute(oid.NameOID.COMMON_NAME, u'bar')]),
])
def test_parse_dn(name, expected):
    result = _parse_dn(name)
    print(result, expected)
    assert result == expected


@pytest.mark.parametrize('name, message', [
    (b'CN=\\0', u'Error while parsing distinguished name "CN=\\0": Hex escape sequence "\\0" incomplete at end of string'),
    (b'CN=x,', u'Error while parsing distinguished name "CN=x,": unexpected end of string'),
])
def test_parse_dn_failure(name, message):
    with pytest.raises(OpenSSLObjectError, match=u'^%s$' % re.escape(message)):
        result = _parse_dn(name)



# (c) 2021 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish

from __future__ import absolute_import, division, print_function

__metaclass__ = type

from ansible_collections.cisco.iosxr.tests.unit.compat.mock import patch
from ansible_collections.cisco.iosxr.plugins.modules import iosxr_l2_interfaces
from ansible_collections.cisco.iosxr.tests.unit.modules.utils import (
    set_module_args,
)
from .iosxr_module import TestIosxrModule, load_fixture


class TestIosxrL2InterfacesModule(TestIosxrModule):
    module = iosxr_l2_interfaces

    def setUp(self):
        super(TestIosxrL2InterfacesModule, self).setUp()

        self.mock_get_config = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.network.Config.get_config"
        )
        self.get_config = self.mock_get_config.start()

        self.mock_load_config = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.network.Config.load_config"
        )
        self.load_config = self.mock_load_config.start()

        self.mock_get_resource_connection_config = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.cfg.base.get_resource_connection"
        )
        self.get_resource_connection_config = (
            self.mock_get_resource_connection_config.start()
        )

        self.mock_get_resource_connection_facts = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.facts.facts.get_resource_connection"
        )
        self.get_resource_connection_facts = (
            self.mock_get_resource_connection_facts.start()
        )
        self.mock_get_os_version = patch(
            "ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.config.l2_interfaces.l2_interfaces.get_os_version"
        )
        self.get_os_version = self.mock_get_os_version.start()
        self.get_os_version.return_value = "7.0.2"
        self.mock_get_os_version1 = patch(
            "ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.facts.l2_interfaces.l2_interfaces.get_os_version"
        )
        self.get_os_version1 = self.mock_get_os_version1.start()
        self.get_os_version1.return_value = "7.0.2"
        self.mock_execute_show_command = patch(
            "ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.facts.l2_interfaces.l2_interfaces.L2_InterfacesFacts.get_config"
        )
        self.execute_show_command = self.mock_execute_show_command.start()

    def tearDown(self):
        super(TestIosxrL2InterfacesModule, self).tearDown()
        self.mock_get_resource_connection_config.stop()
        self.mock_get_resource_connection_facts.stop()
        self.mock_get_config.stop()
        self.mock_get_os_version.stop()
        self.get_os_version1.stop()
        self.mock_load_config.stop()
        self.mock_execute_show_command.stop()

    def _prepare(self):
        def load_from_file(*args, **kwargs):
            return load_fixture("iosxr_l2_interface_config.cfg")

        self.execute_show_command.side_effect = load_from_file

    def test_iosxr_l2_interfaces_merged_idempotent(self):
        self._prepare()
        set_module_args(
            dict(
                config=[
                    dict(
                        name="GigabitEthernet0/0/0/1",
                        l2transport=True,
                        l2protocol=[dict(cpsv="tunnel")],
                    ),
                    dict(
                        name="GigabitEthernet0/0/0/3.900",
                        encapsulation=dict(dot1q=20, second_dot1q=40),
                    ),
                ],
                state="merged",
            )
        )
        self.execute_module(changed=False, commands=[])

    def test_iosxr_l2_interfaces_merged(self):
        set_module_args(
            dict(
                config=[
                    dict(
                        name="GigabitEthernet0/0/0/1",
                        l2transport=True,
                        l2protocol=[dict(cpsv="tunnel")],
                    ),
                    dict(
                        name="GigabitEthernet0/0/0/3.900",
                        encapsulation=dict(dot1q=20, second_dot1q=40),
                    ),
                ],
                state="merged",
            )
        )
        commands = [
            "interface GigabitEthernet0/0/0/1",
            "l2transport l2protocol cpsv tunnel",
            "interface GigabitEthernet0/0/0/3.900",
            "encapsulation dot1q 20 second-dot1q 40",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(sorted(result["commands"]), sorted(commands))

    def test_iosxr_l2_interfaces_replaced(self):
        self._prepare()
        set_module_args(
            dict(
                config=[
                    dict(
                        name="GigabitEthernet0/0/0/1",
                        l2transport=True,
                        l2protocol=[dict(cpsv="drop")],
                    )
                ],
                state="replaced",
            )
        )
        commands = [
            "interface GigabitEthernet0/0/0/1",
            "l2transport l2protocol cpsv drop",
            "no l2transport",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(sorted(result["commands"]), sorted(commands))

    def test_iosxr_l2_interfaces_deleted(self):
        self._prepare()
        set_module_args(dict(state="deleted"))

        commands = [
            "interface GigabitEthernet0/0/0/1",
            "no l2transport",
            "interface GigabitEthernet0/0/0/3.900",
            "no encapsulation dot1q",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(sorted(result["commands"]), sorted(commands))

    def test_iosxr_l2_interfaces_rendered(self):
        set_module_args(
            dict(
                config=[
                    dict(
                        name="GigabitEthernet0/0/0/1",
                        l2transport=True,
                        l2protocol=[dict(cpsv="tunnel")],
                    ),
                    dict(
                        name="GigabitEthernet0/0/0/3.900",
                        encapsulation=dict(dot1q=20, second_dot1q=40),
                    ),
                ],
                state="rendered",
            )
        )

        commands = [
            "interface GigabitEthernet0/0/0/1",
            "l2transport l2protocol cpsv tunnel",
            "interface GigabitEthernet0/0/0/3.900",
            "encapsulation dot1q 20 second-dot1q 40",
        ]
        result = self.execute_module(changed=False)
        self.assertEqual(sorted(result["rendered"]), sorted(commands))

    def test_iosxr_l2_interfaces_parsed(self):
        self.maxDiff = None
        set_module_args(
            dict(
                running_config="interface GigabitEthernet0/0/0/1\n l2transport\n  l2protocol cpsv tunnel\n  "
                "propagate remote-status\n !",
                state="parsed",
            )
        )
        result = self.execute_module(changed=False)
        print(result["parsed"])
        parsed_list = [
            {
                "name": "GigabitEthernet0/0/0/1",
                "l2transport": True,
                "l2protocol": [{"cpsv": "tunnel"}],
                "propagate": True,
            }
        ]
        self.assertEqual(parsed_list, result["parsed"])

    def test_iosxr_l2_interfaces_overridden(self):
        self.maxDiff = None
        self._prepare()
        set_module_args(
            dict(
                config=[
                    dict(
                        name="GigabitEthernet0/0/0/4",
                        l2transport=True,
                        l2protocol=[dict(cpsv="tunnel")],
                    ),
                    dict(
                        name="GigabitEthernet0/0/0/3.900",
                        encapsulation=dict(dot1q=40, second_dot1q=60),
                    ),
                ],
                state="overridden",
            )
        )
        commands = [
            "interface GigabitEthernet0/0/0/4",
            "l2transport l2protocol cpsv tunnel",
            "interface GigabitEthernet0/0/0/3.900",
            "encapsulation dot1q 40 second-dot1q 60",
            "interface GigabitEthernet0/0/0/1",
            "no l2transport",
        ]

        result = self.execute_module(changed=True)
        self.assertEqual(sorted(result["commands"]), sorted(commands))



# -*- coding: utf-8 -*-

# Copyright: (c) 2016 Michael Gruener <michael.gruener@chaosmoon.net>
# Copyright: (c) 2021 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


import base64
import binascii
import datetime
import os
import re
import tempfile
import traceback

from ansible.module_utils.common.text.converters import to_native, to_text, to_bytes

from ansible_collections.community.crypto.plugins.module_utils.acme.backends import (
    CryptoBackend,
)

from ansible_collections.community.crypto.plugins.module_utils.acme.errors import (
    BackendException,
    KeyParsingError,
)

from ansible_collections.community.crypto.plugins.module_utils.acme.utils import nopad_b64

try:
    import ipaddress
except ImportError:
    pass


_OPENSSL_ENVIRONMENT_UPDATE = dict(LANG='C', LC_ALL='C', LC_MESSAGES='C', LC_CTYPE='C')


class OpenSSLCLIBackend(CryptoBackend):
    def __init__(self, module, openssl_binary=None):
        super(OpenSSLCLIBackend, self).__init__(module)
        if openssl_binary is None:
            openssl_binary = module.get_bin_path('openssl', True)
        self.openssl_binary = openssl_binary

    def parse_key(self, key_file=None, key_content=None, passphrase=None):
        '''
        Parses an RSA or Elliptic Curve key file in PEM format and returns key_data.
        Raises KeyParsingError in case of errors.
        '''
        if passphrase is not None:
            raise KeyParsingError('openssl backend does not support key passphrases')
        # If key_file is not given, but key_content, write that to a temporary file
        if key_file is None:
            fd, tmpsrc = tempfile.mkstemp()
            self.module.add_cleanup_file(tmpsrc)  # Ansible will delete the file on exit
            f = os.fdopen(fd, 'wb')
            try:
                f.write(key_content.encode('utf-8'))
                key_file = tmpsrc
            except Exception as err:
                try:
                    f.close()
                except Exception as dummy:
                    pass
                raise KeyParsingError("failed to create temporary content file: %s" % to_native(err), exception=traceback.format_exc())
            f.close()
        # Parse key
        account_key_type = None
        with open(key_file, "rt") as f:
            for line in f:
                m = re.match(r"^\s*-{5,}BEGIN\s+(EC|RSA)\s+PRIVATE\s+KEY-{5,}\s*$", line)
                if m is not None:
                    account_key_type = m.group(1).lower()
                    break
        if account_key_type is None:
            # This happens for example if openssl_privatekey created this key
            # (as opposed to the OpenSSL binary). For now, we assume this is
            # an RSA key.
            # FIXME: add some kind of auto-detection
            account_key_type = "rsa"
        if account_key_type not in ("rsa", "ec"):
            raise KeyParsingError('unknown key type "%s"' % account_key_type)

        openssl_keydump_cmd = [self.openssl_binary, account_key_type, "-in", key_file, "-noout", "-text"]
        dummy, out, dummy = self.module.run_command(
            openssl_keydump_cmd, check_rc=True, environ_update=_OPENSSL_ENVIRONMENT_UPDATE)

        if account_key_type == 'rsa':
            pub_hex, pub_exp = re.search(
                r"modulus:\n\s+00:([a-f0-9\:\s]+?)\npublicExponent: ([0-9]+)",
                to_text(out, errors='surrogate_or_strict'), re.MULTILINE | re.DOTALL).groups()
            pub_exp = "{0:x}".format(int(pub_exp))
            if len(pub_exp) % 2:
                pub_exp = "0{0}".format(pub_exp)

            return {
                'key_file': key_file,
                'type': 'rsa',
                'alg': 'RS256',
                'jwk': {
                    "kty": "RSA",
                    "e": nopad_b64(binascii.unhexlify(pub_exp.encode("utf-8"))),
                    "n": nopad_b64(binascii.unhexlify(re.sub(r"(\s|:)", "", pub_hex).encode("utf-8"))),
                },
                'hash': 'sha256',
            }
        elif account_key_type == 'ec':
            pub_data = re.search(
                r"pub:\s*\n\s+04:([a-f0-9\:\s]+?)\nASN1 OID: (\S+)(?:\nNIST CURVE: (\S+))?",
                to_text(out, errors='surrogate_or_strict'), re.MULTILINE | re.DOTALL)
            if pub_data is None:
                raise KeyParsingError('cannot parse elliptic curve key')
            pub_hex = binascii.unhexlify(re.sub(r"(\s|:)", "", pub_data.group(1)).encode("utf-8"))
            asn1_oid_curve = pub_data.group(2).lower()
            nist_curve = pub_data.group(3).lower() if pub_data.group(3) else None
            if asn1_oid_curve == 'prime256v1' or nist_curve == 'p-256':
                bits = 256
                alg = 'ES256'
                hashalg = 'sha256'
                point_size = 32
                curve = 'P-256'
            elif asn1_oid_curve == 'secp384r1' or nist_curve == 'p-384':
                bits = 384
                alg = 'ES384'
                hashalg = 'sha384'
                point_size = 48
                curve = 'P-384'
            elif asn1_oid_curve == 'secp521r1' or nist_curve == 'p-521':
                # Not yet supported on Let's Encrypt side, see
                # https://github.com/letsencrypt/boulder/issues/2217
                bits = 521
                alg = 'ES512'
                hashalg = 'sha512'
                point_size = 66
                curve = 'P-521'
            else:
                raise KeyParsingError('unknown elliptic curve: %s / %s' % (asn1_oid_curve, nist_curve))
            num_bytes = (bits + 7) // 8
            if len(pub_hex) != 2 * num_bytes:
                raise KeyParsingError('bad elliptic curve point (%s / %s)' % (asn1_oid_curve, nist_curve))
            return {
                'key_file': key_file,
                'type': 'ec',
                'alg': alg,
                'jwk': {
                    "kty": "EC",
                    "crv": curve,
                    "x": nopad_b64(pub_hex[:num_bytes]),
                    "y": nopad_b64(pub_hex[num_bytes:]),
                },
                'hash': hashalg,
                'point_size': point_size,
            }

    def sign(self, payload64, protected64, key_data):
        sign_payload = "{0}.{1}".format(protected64, payload64).encode('utf8')
        if key_data['type'] == 'hmac':
            hex_key = to_native(binascii.hexlify(base64.urlsafe_b64decode(key_data['jwk']['k'])))
            cmd_postfix = ["-mac", "hmac", "-macopt", "hexkey:{0}".format(hex_key), "-binary"]
        else:
            cmd_postfix = ["-sign", key_data['key_file']]
        openssl_sign_cmd = [self.openssl_binary, "dgst", "-{0}".format(key_data['hash'])] + cmd_postfix

        dummy, out, dummy = self.module.run_command(
            openssl_sign_cmd, data=sign_payload, check_rc=True, binary_data=True, environ_update=_OPENSSL_ENVIRONMENT_UPDATE)

        if key_data['type'] == 'ec':
            dummy, der_out, dummy = self.module.run_command(
                [self.openssl_binary, "asn1parse", "-inform", "DER"],
                data=out, binary_data=True, environ_update=_OPENSSL_ENVIRONMENT_UPDATE)
            expected_len = 2 * key_data['point_size']
            sig = re.findall(
                r"prim:\s+INTEGER\s+:([0-9A-F]{1,%s})\n" % expected_len,
                to_text(der_out, errors='surrogate_or_strict'))
            if len(sig) != 2:
                raise BackendException(
                    "failed to generate Elliptic Curve signature; cannot parse DER output: {0}".format(
                        to_text(der_out, errors='surrogate_or_strict')))
            sig[0] = (expected_len - len(sig[0])) * '0' + sig[0]
            sig[1] = (expected_len - len(sig[1])) * '0' + sig[1]
            out = binascii.unhexlify(sig[0]) + binascii.unhexlify(sig[1])

        return {
            "protected": protected64,
            "payload": payload64,
            "signature": nopad_b64(to_bytes(out)),
        }

    def create_mac_key(self, alg, key):
        '''Create a MAC key.'''
        if alg == 'HS256':
            hashalg = 'sha256'
            hashbytes = 32
        elif alg == 'HS384':
            hashalg = 'sha384'
            hashbytes = 48
        elif alg == 'HS512':
            hashalg = 'sha512'
            hashbytes = 64
        else:
            raise BackendException('Unsupported MAC key algorithm for OpenSSL backend: {0}'.format(alg))
        key_bytes = base64.urlsafe_b64decode(key)
        if len(key_bytes) < hashbytes:
            raise BackendException(
                '{0} key must be at least {1} bytes long (after Base64 decoding)'.format(alg, hashbytes))
        return {
            'type': 'hmac',
            'alg': alg,
            'jwk': {
                'kty': 'oct',
                'k': key,
            },
            'hash': hashalg,
        }

    @staticmethod
    def _normalize_ip(ip):
        try:
            return to_native(ipaddress.ip_address(to_text(ip)).compressed)
        except ValueError:
            # We do not want to error out on something IPAddress() cannot parse
            return ip

    def get_csr_identifiers(self, csr_filename=None, csr_content=None):
        '''
        Return a set of requested identifiers (CN and SANs) for the CSR.
        Each identifier is a pair (type, identifier), where type is either
        'dns' or 'ip'.
        '''
        filename = csr_filename
        data = None
        if csr_content is not None:
            filename = '/dev/stdin'
            data = csr_content.encode('utf-8')

        openssl_csr_cmd = [self.openssl_binary, "req", "-in", filename, "-noout", "-text"]
        dummy, out, dummy = self.module.run_command(
            openssl_csr_cmd, data=data, check_rc=True, binary_data=True, environ_update=_OPENSSL_ENVIRONMENT_UPDATE)

        identifiers = set([])
        common_name = re.search(r"Subject:.* CN\s?=\s?([^\s,;/]+)", to_text(out, errors='surrogate_or_strict'))
        if common_name is not None:
            identifiers.add(('dns', common_name.group(1)))
        subject_alt_names = re.search(
            r"X509v3 Subject Alternative Name: (?:critical)?\n +([^\n]+)\n",
            to_text(out, errors='surrogate_or_strict'), re.MULTILINE | re.DOTALL)
        if subject_alt_names is not None:
            for san in subject_alt_names.group(1).split(", "):
                if san.lower().startswith("dns:"):
                    identifiers.add(('dns', san[4:]))
                elif san.lower().startswith("ip:"):
                    identifiers.add(('ip', self._normalize_ip(san[3:])))
                elif san.lower().startswith("ip address:"):
                    identifiers.add(('ip', self._normalize_ip(san[11:])))
                else:
                    raise BackendException('Found unsupported SAN identifier "{0}"'.format(san))
        return identifiers

    def get_cert_days(self, cert_filename=None, cert_content=None, now=None):
        '''
        Return the days the certificate in cert_filename remains valid and -1
        if the file was not found. If cert_filename contains more than one
        certificate, only the first one will be considered.

        If now is not specified, datetime.datetime.now() is used.
        '''
        filename = cert_filename
        data = None
        if cert_content is not None:
            filename = '/dev/stdin'
            data = cert_content.encode('utf-8')
            cert_filename_suffix = ''
        elif cert_filename is not None:
            if not os.path.exists(cert_filename):
                return -1
            cert_filename_suffix = ' in {0}'.format(cert_filename)
        else:
            return -1

        openssl_cert_cmd = [self.openssl_binary, "x509", "-in", filename, "-noout", "-text"]
        dummy, out, dummy = self.module.run_command(
            openssl_cert_cmd, data=data, check_rc=True, binary_data=True, environ_update=_OPENSSL_ENVIRONMENT_UPDATE)
        try:
            not_after_str = re.search(r"\s+Not After\s*:\s+(.*)", to_text(out, errors='surrogate_or_strict')).group(1)
            not_after = datetime.datetime.strptime(not_after_str, '%b %d %H:%M:%S %Y %Z')
        except AttributeError:
            raise BackendException("No 'Not after' date found{0}".format(cert_filename_suffix))
        except ValueError:
            raise BackendException("Failed to parse 'Not after' date{0}".format(cert_filename_suffix))
        if now is None:
            now = datetime.datetime.now()
        return (not_after - now).days

    def create_chain_matcher(self, criterium):
        '''
        Given a Criterium object, creates a ChainMatcher object.
        '''
        raise BackendException('Alternate chain matching can only be used with the "cryptography" backend.')



#
# (c) 2016 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, division, print_function

__metaclass__ = type

import re
import json

from ansible.errors import AnsibleConnectionFailure
from ansible.module_utils._text import to_text, to_bytes
from ansible_collections.ansible.netcommon.plugins.plugin_utils.terminal_base import (
    TerminalBase,
)


class TerminalModule(TerminalBase):

    terminal_stdout_re = [
        re.compile(rb"[\r\n]?[\w+\-\.:\/\[\]]+(?:\([^\)]+\)){,3}(?:>|#) ?$"),
        re.compile(rb"\[\w+\@[\w\-\.]+(?: [^\]])\] ?[>#\$] ?$"),
    ]

    terminal_stderr_re = [
        re.compile(rb"error:", re.I),
        re.compile(rb"Removing.* not allowed, it is being used"),
        re.compile(rb"^Command authorization failed\r?$", re.MULTILINE),
    ]

    terminal_config_prompt = re.compile(r"^.+\(config(-.*)?\)#$")

    def on_open_shell(self):
        if self._get_prompt().strip().endswith(b"#"):
            self.disable_pager()

    def disable_pager(self):
        try:
            self._exec_cli_command("no terminal pager")
        except AnsibleConnectionFailure:
            raise AnsibleConnectionFailure("unable to disable terminal pager")

    def on_become(self, passwd=None):
        if self._get_prompt().strip().endswith(b"#"):
            return

        cmd = {"command": "enable"}
        if passwd:
            # Note: python-3.5 cannot combine u"" and r"" together.  Thus make
            # an r string and use to_text to ensure it's text on both py2 and py3.
            cmd["prompt"] = to_text(
                r"[\r\n]?[Pp]assword: $", errors="surrogate_or_strict"
            )
            cmd["answer"] = passwd

        try:
            self._exec_cli_command(
                to_bytes(json.dumps(cmd), errors="surrogate_or_strict")
            )
        except AnsibleConnectionFailure:
            raise AnsibleConnectionFailure(
                "unable to elevate privilege to enable mode"
            )

        self.disable_pager()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: sensor_info
short_description: Information module for Sensor
description:
- Get all Sensor.
- Intent API to get a list of SENSOR devices.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module_info
author: Rafael Campos (@racampos)
options:
  headers:
    description: Additional headers.
    type: dict
  siteId:
    description:
    - SiteId query parameter.
    type: str
requirements:
- dnacentersdk >= 2.5.0
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Sensors Sensors
  description: Complete reference of the Sensors API.
  link: https://developer.cisco.com/docs/dna-center/#!sensors-sensors
notes:
  - SDK Method used are
    sensors.Sensors.sensors,

  - Paths used are
    get /dna/intent/api/v1/sensor,

"""

EXAMPLES = r"""
- name: Get all Sensor
  cisco.dnac.sensor_info:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    headers: "{{my_headers | from_json}}"
    siteId: string
  register: result

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "version": "string",
      "response": [
        {
          "name": "string",
          "status": "string",
          "radioMacAddress": "string",
          "ethernetMacAddress": "string",
          "location": "string",
          "backhaulType": "string",
          "serialNumber": "string",
          "ipAddress": "string",
          "version": "string",
          "lastSeen": 0,
          "type": "string",
          "sshConfig": {
            "sshState": "string",
            "sshUserName": "string",
            "sshPassword": "string",
            "enablePassword": "string"
          },
          "isLEDEnabled": true
        }
      ]
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: network_access_dictionary_attributes_authentication_info
short_description: Information module for Network Access Dictionary Attributes Authentication
description:
- Get all Network Access Dictionary Attributes Authentication.
- Network Access - Returns list of dictionary attributes for authentication.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.ise.module_info
author: Rafael Campos (@racampos)
options: {}
requirements:
- ciscoisesdk >= 2.0.3
- python >= 3.5
seealso:
- name: Cisco ISE documentation for Network Access - Dictionary Attributes List
  description: Complete reference of the Network Access - Dictionary Attributes List API.
  link: https://developer.cisco.com/docs/identity-services-engine/v1/#!policy-openapi
notes:
  - SDK Method used are
    network_access_dictionary_attributes_list.NetworkAccessDictionaryAttributesList.get_network_access_dictionaries_authentication,

  - Paths used are
    get /network-access/dictionaries/authentication,

"""

EXAMPLES = r"""
- name: Get all Network Access Dictionary Attributes Authentication
  cisco.ise.network_access_dictionary_attributes_authentication_info:
    ise_hostname: "{{ise_hostname}}"
    ise_username: "{{ise_username}}"
    ise_password: "{{ise_password}}"
    ise_verify: "{{ise_verify}}"
  register: result

"""

RETURN = r"""
ise_response:
  description: A dictionary or list with the response returned by the Cisco ISE Python SDK
  returned: always
  type: list
  elements: dict
  sample: >
    [
      {
        "allowedValues": [
          {
            "isDefault": true,
            "key": "string",
            "value": "string"
          }
        ],
        "dataType": "string",
        "description": "string",
        "dictionaryName": "string",
        "directionType": "string",
        "id": "string",
        "internalName": "string",
        "name": "string"
      }
    ]
"""



# -*- coding: utf-8 -*-
# Copyright (c) 2021, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Make coding more python3-ish
from __future__ import absolute_import, division, print_function

__metaclass__ = type


import pytest

from ansible_collections.community.dns.plugins.module_utils.conversion.base import (
    DNSConversionError,
)

from ansible_collections.community.dns.plugins.module_utils.conversion.converter import (
    RecordConverter,
)

from ansible_collections.community.dns.plugins.module_utils.record import (
    DNSRecord,
)

from ..helper import (
    CustomProviderInformation,
    CustomProvideOptions,
)


def test_user_api():
    converter = RecordConverter(
        CustomProviderInformation(txt_record_handling='decoded'),
        CustomProvideOptions({'txt_transformation': 'api'}))
    assert converter.process_value_from_user('TXT', u'"xyz \\') == u'"xyz \\'
    assert converter.process_values_from_user('TXT', [u'"xyz \\']) == [u'"xyz \\']
    assert converter.process_value_to_user('TXT', u'"xyz \\') == u'"xyz \\'
    assert converter.process_values_to_user('TXT', [u'"xyz \\']) == [u'"xyz \\']

    record = DNSRecord()
    record.type = 'TXT'

    record.target = u'"xyz \\'
    converter.process_from_user(record)
    assert record.target == u'"xyz \\'

    record.target = u'"xyz \\'
    converter.process_multiple_from_user([record])
    assert record.target == u'"xyz \\'

    record.target = u'"xyz \\'
    converter.process_to_user(record)
    assert record.target == u'"xyz \\'

    record.target = u'"xyz \\'
    converter.process_multiple_to_user([record])
    assert record.target == u'"xyz \\'


def test_user_quoted():
    converter = RecordConverter(
        CustomProviderInformation(txt_record_handling='decoded'),
        CustomProvideOptions({'txt_transformation': 'quoted'}))
    assert converter.process_value_from_user('TXT', u'hëllo " w\\303\\266rld"') == u'hëllo wörld'
    assert converter.process_values_from_user('TXT', [u'hëllo " w\\303\\266rld"']) == [u'hëllo wörld']
    assert converter.process_value_to_user('TXT', u'hello wörld') == u'"hello w\\303\\266rld"'
    assert converter.process_values_to_user('TXT', [u'hello wörld']) == [u'"hello w\\303\\266rld"']

    record = DNSRecord()
    record.type = 'TXT'

    record.target = u'hëllo " w\\303\\266rld"'
    converter.process_from_user(record)
    assert record.target == u'hëllo wörld'

    record.target = u'hëllo " w\\303\\266rld"'
    converter.process_multiple_from_user([record])
    assert record.target == u'hëllo wörld'

    record.target = u'hello wörld'
    converter.process_to_user(record)
    assert record.target == u'"hello w\\303\\266rld"'

    record.target = u'hello wörld'
    converter.process_multiple_to_user([record])
    assert record.target == u'"hello w\\303\\266rld"'

    record.target = u'"a\\o'
    with pytest.raises(DNSConversionError) as exc:
        converter.process_from_user(record)
    print(exc.value.error_message)
    assert exc.value.error_message == (
        u'While processing record from the user: A backslash must not be followed by "o" (index 4)'
    )


def test_user_unquoted():
    converter = RecordConverter(
        CustomProviderInformation(txt_record_handling='decoded'),
        CustomProvideOptions({'txt_transformation': 'unquoted'}))
    assert converter.process_value_from_user('TXT', u'hello "wörl\\d"') == u'hello "wörl\\d"'
    assert converter.process_values_from_user('TXT', [u'hello "wörl\\d"']) == [u'hello "wörl\\d"']
    assert converter.process_value_to_user('TXT', u'hello "wörl\\d"') == u'hello "wörl\\d"'
    assert converter.process_values_to_user('TXT', [u'hello "wörl\\d"']) == [u'hello "wörl\\d"']

    record = DNSRecord()
    record.type = 'TXT'

    record.target = u'hello "wörl\\d"'
    converter.process_from_user(record)
    assert record.target == u'hello "wörl\\d"'

    record.target = u'hello "wörl\\d"'
    converter.process_multiple_from_user([record])
    assert record.target == u'hello "wörl\\d"'

    record.target = u'hello "wörl\\d"'
    converter.process_to_user(record)
    assert record.target == u'hello "wörl\\d"'

    record.target = u'hello "wörl\\d"'
    converter.process_multiple_to_user([record])
    assert record.target == u'hello "wörl\\d"'


def test_api_decoded():
    converter = RecordConverter(
        CustomProviderInformation(txt_record_handling='decoded'),
        CustomProvideOptions({'txt_transformation': 'unquoted'}))
    record = DNSRecord()
    record.type = 'TXT'

    record.target = u'"xyz \\'
    record_2 = converter.clone_from_api(record)
    assert record is not record_2
    assert record.target == u'"xyz \\'
    assert record_2.target == u'"xyz \\'
    converter.process_from_api(record)
    assert record.target == u'"xyz \\'

    record.target = u'"xyz \\'
    records = converter.clone_multiple_from_api([record])
    assert len(records) == 1
    assert record is not records[0]
    assert record.target == u'"xyz \\'
    assert records[0].target == u'"xyz \\'
    converter.process_multiple_from_api([record])
    assert record.target == u'"xyz \\'

    record.target = u'"xyz \\'
    record_2 = converter.clone_to_api(record)
    assert record is not record_2
    assert record.target == u'"xyz \\'
    assert record_2.target == u'"xyz \\'
    converter.process_to_api(record)
    assert record.target == u'"xyz \\'

    record.target = u'"xyz \\'
    records = converter.clone_multiple_to_api([record])
    assert len(records) == 1
    assert record is not records[0]
    assert record.target == u'"xyz \\'
    assert records[0].target == u'"xyz \\'
    converter.process_multiple_to_api([record])
    assert record.target == u'"xyz \\'


def test_api_encoded():
    converter = RecordConverter(
        CustomProviderInformation(txt_record_handling='encoded'),
        CustomProvideOptions({'txt_transformation': 'unquoted'}))
    record = DNSRecord()
    record.type = 'TXT'

    record.target = u'xyz " " \\\\\\303\\266'
    record_2 = converter.clone_from_api(record)
    assert record is not record_2
    assert record.target == u'xyz " " \\\\\\303\\266'
    print(record_2.target)
    assert record_2.target == u'xyz \\ö'
    converter.process_from_api(record)
    assert record.target == u'xyz \\ö'

    record.target = u'xyz " " \\\\\\303\\266'
    records = converter.clone_multiple_from_api([record])
    assert len(records) == 1
    assert record is not records[0]
    assert record.target == u'xyz " " \\\\\\303\\266'
    assert records[0].target == u'xyz \\ö'
    converter.process_multiple_from_api([record])
    assert record.target == u'xyz \\ö'

    record.target = u'xyz \\ö'
    record_2 = converter.clone_to_api(record)
    assert record is not record_2
    assert record.target == u'xyz \\ö'
    assert record_2.target == u'"xyz \\\\\\303\\266"'
    converter.process_to_api(record)
    assert record.target == u'"xyz \\\\\\303\\266"'

    record.target = u'xyz \\ö'
    records = converter.clone_multiple_to_api([record])
    assert len(records) == 1
    assert record is not records[0]
    assert record.target == u'xyz \\ö'
    assert records[0].target == u'"xyz \\\\\\303\\266"'
    converter.process_multiple_to_api([record])
    assert record.target == u'"xyz \\\\\\303\\266"'

    record.target = u'"a'
    with pytest.raises(DNSConversionError) as exc:
        converter.process_from_api(record)
    print(exc.value.error_message)
    assert exc.value.error_message == (
        u'While processing record from API: Missing double quotation mark at the end of value'
    )


def test_api_encoded_no_octal():
    converter = RecordConverter(
        CustomProviderInformation(txt_record_handling='encoded-no-octal'),
        CustomProvideOptions({'txt_transformation': 'unquoted'}))
    record = DNSRecord()
    record.type = 'TXT'

    record.target = u'xyz " " \\\\\\303\\266'
    record_2 = converter.clone_from_api(record)
    assert record is not record_2
    assert record.target == u'xyz " " \\\\\\303\\266'
    print(record_2.target)
    assert record_2.target == u'xyz \\ö'
    converter.process_from_api(record)
    assert record.target == u'xyz \\ö'

    record.target = u'xyz " " \\\\\\303\\266'
    records = converter.clone_multiple_from_api([record])
    assert len(records) == 1
    assert record is not records[0]
    assert record.target == u'xyz " " \\\\\\303\\266'
    assert records[0].target == u'xyz \\ö'
    converter.process_multiple_from_api([record])
    assert record.target == u'xyz \\ö'

    record.target = u'xyz \\ö"'
    record_2 = converter.clone_to_api(record)
    assert record is not record_2
    assert record.target == u'xyz \\ö"'
    assert record_2.target == u'"xyz \\\\ö\\""'
    converter.process_to_api(record)
    assert record.target == u'"xyz \\\\ö\\""'

    record.target = u'xyz \\ö"'
    records = converter.clone_multiple_to_api([record])
    assert len(records) == 1
    assert record is not records[0]
    assert record.target == u'xyz \\ö"'
    assert records[0].target == u'"xyz \\\\ö\\""'
    converter.process_multiple_to_api([record])
    assert record.target == u'"xyz \\\\ö\\""'



#!/usr/bin/python
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


DOCUMENTATION = '''
---
module: ec2_vpc_egress_igw
version_added: 1.0.0
short_description: Manage an AWS VPC Egress Only Internet gateway
description:
    - Manage an AWS VPC Egress Only Internet gateway
author: Daniel Shepherd (@shepdelacreme)
options:
  vpc_id:
    description:
      - The VPC ID for the VPC that this Egress Only Internet Gateway should be attached.
    required: true
    type: str
  state:
    description:
      - Create or delete the EIGW.
    default: present
    choices: [ 'present', 'absent' ]
    type: str
extends_documentation_fragment:
- amazon.aws.aws
- amazon.aws.ec2

'''

EXAMPLES = '''
# Note: These examples do not set authentication details, see the AWS Guide for details.

# Ensure that the VPC has an Internet Gateway.
# The Internet Gateway ID is can be accessed via {{eigw.gateway_id}} for use in setting up NATs etc.
- community.aws.ec2_vpc_egress_igw:
    vpc_id: vpc-abcdefgh
    state: present
  register: eigw

'''

RETURN = '''
gateway_id:
    description: The ID of the Egress Only Internet Gateway or Null.
    returned: always
    type: str
    sample: eigw-0e00cf111ba5bc11e
vpc_id:
    description: The ID of the VPC to attach or detach gateway from.
    returned: always
    type: str
    sample: vpc-012345678
'''

try:
    import botocore
except ImportError:
    pass  # caught by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.core import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.core import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AWSRetry


def delete_eigw(module, connection, eigw_id):
    """
    Delete EIGW.

    module     : AnsibleAWSModule object
    connection : boto3 client connection object
    eigw_id    : ID of the EIGW to delete
    """
    changed = False

    try:
        response = connection.delete_egress_only_internet_gateway(
            aws_retry=True,
            DryRun=module.check_mode,
            EgressOnlyInternetGatewayId=eigw_id)
    except is_boto3_error_code('DryRunOperation'):
        changed = True
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg="Could not delete Egress-Only Internet Gateway {0} from VPC {1}".format(eigw_id, module.vpc_id))

    if not module.check_mode:
        changed = response.get('ReturnCode', False)

    return changed


def create_eigw(module, connection, vpc_id):
    """
    Create EIGW.

    module       : AnsibleAWSModule object
    connection   : boto3 client connection object
    vpc_id       : ID of the VPC we are operating on
    """
    gateway_id = None
    changed = False

    try:
        response = connection.create_egress_only_internet_gateway(
            aws_retry=True,
            DryRun=module.check_mode,
            VpcId=vpc_id)
    except is_boto3_error_code('DryRunOperation'):
        # When boto3 method is run with DryRun=True it returns an error on success
        # We need to catch the error and return something valid
        changed = True
    except is_boto3_error_code('InvalidVpcID.NotFound') as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg="invalid vpc ID '{0}' provided".format(vpc_id))
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg="Could not create Egress-Only Internet Gateway for vpc ID {0}".format(vpc_id))

    if not module.check_mode:
        gateway = response.get('EgressOnlyInternetGateway', {})
        state = gateway.get('Attachments', [{}])[0].get('State')
        gateway_id = gateway.get('EgressOnlyInternetGatewayId')

        if gateway_id and state in ('attached', 'attaching'):
            changed = True
        else:
            # EIGW gave back a bad attachment state or an invalid response so we error out
            module.fail_json(msg='Unable to create and attach Egress Only Internet Gateway to VPCId: {0}. Bad or no state in response'.format(vpc_id),
                             **camel_dict_to_snake_dict(response))

    return changed, gateway_id


def describe_eigws(module, connection, vpc_id):
    """
    Describe EIGWs.

    module     : AnsibleAWSModule object
    connection : boto3 client connection object
    vpc_id     : ID of the VPC we are operating on
    """
    gateway_id = None

    try:
        response = connection.describe_egress_only_internet_gateways(
            aws_retry=True)
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg="Could not get list of existing Egress-Only Internet Gateways")

    for eigw in response.get('EgressOnlyInternetGateways', []):
        for attachment in eigw.get('Attachments', []):
            if attachment.get('VpcId') == vpc_id and attachment.get('State') in ('attached', 'attaching'):
                gateway_id = eigw.get('EgressOnlyInternetGatewayId')

    return gateway_id


def main():
    argument_spec = dict(
        vpc_id=dict(required=True),
        state=dict(default='present', choices=['present', 'absent'])
    )

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True)

    retry_decorator = AWSRetry.jittered_backoff(retries=10)
    connection = module.client('ec2', retry_decorator=retry_decorator)

    vpc_id = module.params.get('vpc_id')
    state = module.params.get('state')

    eigw_id = describe_eigws(module, connection, vpc_id)

    result = dict(gateway_id=eigw_id, vpc_id=vpc_id)
    changed = False

    if state == 'present' and not eigw_id:
        changed, result['gateway_id'] = create_eigw(module, connection, vpc_id)
    elif state == 'absent' and eigw_id:
        changed = delete_eigw(module, connection, eigw_id)

    module.exit_json(changed=changed, **result)


if __name__ == '__main__':
    main()



from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

import re
import copy

try:
    import botocore
except ImportError:
    pass  # handled by AnsibleAWSModule

from ansible_collections.amazon.aws.plugins.module_utils.core import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import camel_dict_to_snake_dict


@AWSRetry.jittered_backoff()
def _list_topics_with_backoff(client):
    paginator = client.get_paginator('list_topics')
    return paginator.paginate().build_full_result()['Topics']


@AWSRetry.jittered_backoff(catch_extra_error_codes=['NotFound'])
def _list_topic_subscriptions_with_backoff(client, topic_arn):
    paginator = client.get_paginator('list_subscriptions_by_topic')
    return paginator.paginate(TopicArn=topic_arn).build_full_result()['Subscriptions']


@AWSRetry.jittered_backoff(catch_extra_error_codes=['NotFound'])
def _list_subscriptions_with_backoff(client):
    paginator = client.get_paginator('list_subscriptions')
    return paginator.paginate().build_full_result()['Subscriptions']


def list_topic_subscriptions(client, module, topic_arn):
    try:
        return _list_topic_subscriptions_with_backoff(client, topic_arn)
    except is_boto3_error_code('AuthorizationError'):
        try:
            # potentially AuthorizationError when listing subscriptions for third party topic
            return [sub for sub in _list_subscriptions_with_backoff(client)
                    if sub['TopicArn'] == topic_arn]
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            module.fail_json_aws(e, msg="Couldn't get subscriptions list for topic %s" % topic_arn)
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg="Couldn't get subscriptions list for topic %s" % topic_arn)


def list_topics(client, module):
    try:
        topics = _list_topics_with_backoff(client)
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Couldn't get topic list")
    return [t['TopicArn'] for t in topics]


def topic_arn_lookup(client, module, name):
    # topic names cannot have colons, so this captures the full topic name
    all_topics = list_topics(client, module)
    lookup_topic = ':%s' % name
    for topic in all_topics:
        if topic.endswith(lookup_topic):
            return topic


def compare_delivery_policies(policy_a, policy_b):
    _policy_a = copy.deepcopy(policy_a)
    _policy_b = copy.deepcopy(policy_b)
    # AWS automatically injects disableSubscriptionOverrides if you set an
    # http policy
    if 'http' in policy_a:
        if 'disableSubscriptionOverrides' not in policy_a['http']:
            _policy_a['http']['disableSubscriptionOverrides'] = False
    if 'http' in policy_b:
        if 'disableSubscriptionOverrides' not in policy_b['http']:
            _policy_b['http']['disableSubscriptionOverrides'] = False
    comparison = (_policy_a != _policy_b)
    return comparison


def canonicalize_endpoint(protocol, endpoint):
    # AWS SNS expects phone numbers in
    # and canonicalizes to E.164 format
    # See <https://docs.aws.amazon.com/sns/latest/dg/sms_publish-to-phone.html>
    if protocol == 'sms':
        return re.sub('[^0-9+]*', '', endpoint)
    return endpoint


def get_info(connection, module, topic_arn):
    name = module.params.get('name')
    topic_type = module.params.get('topic_type')
    state = module.params.get('state')
    subscriptions = module.params.get('subscriptions')
    purge_subscriptions = module.params.get('purge_subscriptions')
    subscriptions_existing = module.params.get('subscriptions_existing', [])
    subscriptions_deleted = module.params.get('subscriptions_deleted', [])
    subscriptions_added = module.params.get('subscriptions_added', [])
    subscriptions_added = module.params.get('subscriptions_added', [])
    topic_created = module.params.get('topic_created', False)
    topic_deleted = module.params.get('topic_deleted', False)
    attributes_set = module.params.get('attributes_set', [])
    check_mode = module.check_mode

    info = {
        'name': name,
        'topic_type': topic_type,
        'state': state,
        'subscriptions_new': subscriptions,
        'subscriptions_existing': subscriptions_existing,
        'subscriptions_deleted': subscriptions_deleted,
        'subscriptions_added': subscriptions_added,
        'subscriptions_purge': purge_subscriptions,
        'check_mode': check_mode,
        'topic_created': topic_created,
        'topic_deleted': topic_deleted,
        'attributes_set': attributes_set,
    }
    if state != 'absent':
        if topic_arn in list_topics(connection, module):
            info.update(camel_dict_to_snake_dict(connection.get_topic_attributes(TopicArn=topic_arn)['Attributes']))
            info['delivery_policy'] = info.pop('effective_delivery_policy')
        info['subscriptions'] = [camel_dict_to_snake_dict(sub) for sub in list_topic_subscriptions(connection, module, topic_arn)]

    return info



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type
from ansible.plugins.action import ActionBase

try:
    from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
        AnsibleArgSpecValidator,
    )
except ImportError:
    ANSIBLE_UTILS_IS_INSTALLED = False
else:
    ANSIBLE_UTILS_IS_INSTALLED = True
from ansible.errors import AnsibleActionFail
from ansible_collections.cisco.ise.plugins.plugin_utils.ise import (
    ISESDK,
    ise_argument_spec,
)

# Get common arguements specification
argument_spec = ise_argument_spec()
# Add arguments specific for this module
argument_spec.update(dict(
    id=dict(type="str"),
    page=dict(type="int"),
    size=dict(type="int"),
    filter=dict(type="list"),
    filterType=dict(type="str"),
))

required_if = []
required_one_of = []
mutually_exclusive = []
required_together = []


class ActionModule(ActionBase):
    def __init__(self, *args, **kwargs):
        if not ANSIBLE_UTILS_IS_INSTALLED:
            raise AnsibleActionFail("ansible.utils is not installed. Execute 'ansible-galaxy collection install ansible.utils'")
        super(ActionModule, self).__init__(*args, **kwargs)
        self._supports_async = False
        self._supports_check_mode = True
        self._result = None

    # Checks the supplied parameters against the argument spec for this module
    def _check_argspec(self):
        aav = AnsibleArgSpecValidator(
            data=self._task.args,
            schema=dict(argument_spec=argument_spec),
            schema_format="argspec",
            schema_conditionals=dict(
                required_if=required_if,
                required_one_of=required_one_of,
                mutually_exclusive=mutually_exclusive,
                required_together=required_together,
            ),
            name=self._task.action,
        )
        valid, errors, self._task.args = aav.validate()
        if not valid:
            raise AnsibleActionFail(errors)

    def get_object(self, params):
        new_object = dict(
            id=params.get("id"),
            page=params.get("page"),
            size=params.get("size"),
            filter=params.get("filter"),
            filter_type=params.get("filterType"),
        )
        return new_object

    def run(self, tmp=None, task_vars=None):
        self._task.diff = False
        self._result = super(ActionModule, self).run(tmp, task_vars)
        self._result["changed"] = False
        self._check_argspec()

        self._result.update(dict(ise_response=[]))

        ise = ISESDK(params=self._task.args)

        id = self._task.args.get("id")
        name = self._task.args.get("name")
        if id:
            response = ise.exec(
                family="telemetry_information",
                function='get_telemetry_info_by_id',
                params=self.get_object(self._task.args)
            ).response['TelemetryInfo']
            self._result.update(dict(ise_response=response))
            self._result.update(ise.exit_json())
            return self._result
        if not name and not id:
            responses = []
            generator = ise.exec(
                family="telemetry_information",
                function='get_telemetry_information_generator',
                params=self.get_object(self._task.args),
            )
            try:
                for item in generator:
                    tmp_response = item.response['SearchResult']['resources']
                    if isinstance(tmp_response, list):
                        responses += tmp_response
                    else:
                        responses.append(tmp_response)
                response = responses
            except (TypeError, AttributeError) as e:
                ise.fail_json(
                    msg=(
                        "An error occured when executing operation."
                        " Check the configuration of your API Settings and API Gateway settings on your ISE server."
                        " This collection assumes that the API Gateway, the ERS APIs and OpenAPIs are enabled."
                        " You may want to enable the (ise_debug: True) argument."
                        " The error was: {error}"
                    ).format(error=e)
                )
            except Exception as e:
                ise.fail_json(
                    msg=(
                        "An error occured when executing operation."
                        " The error was: {error}"
                        " You may want to enable the (ise_debug: True) argument."
                    ).format(error=e)
                )

            self._result.update(dict(ise_response=response))
            self._result.update(ise.exit_json())
            return self._result



# (c) 2021 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

# Make coding more python3-ish

from __future__ import absolute_import, division, print_function

__metaclass__ = type

from textwrap import dedent
from ansible_collections.cisco.iosxr.tests.unit.compat.mock import patch
from ansible_collections.cisco.iosxr.plugins.modules import iosxr_hostname
from ansible_collections.cisco.iosxr.tests.unit.modules.utils import (
    set_module_args,
)
from .iosxr_module import TestIosxrModule


class TestIosxrHostnameModule(TestIosxrModule):
    module = iosxr_hostname

    def setUp(self):
        super(TestIosxrHostnameModule, self).setUp()

        self.mock_get_resource_connection = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module_base."
            "get_resource_connection"
        )
        self.get_resource_connection = (
            self.mock_get_resource_connection.start()
        )

        self.mock_get_config = patch(
            "ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.facts.hostname.hostname."
            "HostnameFacts.get_config"
        )
        self.get_config = self.mock_get_config.start()

    def tearDown(self):
        super(TestIosxrHostnameModule, self).tearDown()
        self.get_resource_connection.stop()
        self.get_config.stop()

    def test_iosxr_hostname_merged_idempotent(self):
        self.maxDiff = None
        run_cfg = dedent(
            """\
                hostname iosxr1
            """
        )
        self.get_config.return_value = run_cfg
        set_module_args(dict(config=dict(hostname="iosxr1"), state="merged"))
        self.execute_module(changed=False, commands=[])

    def test_iosxr_hostname_merged(self):
        self.maxDiff = None
        set_module_args(dict(config=dict(hostname="iosxr1"), state="merged"))
        commands = ["hostname iosxr1"]
        result = self.execute_module(changed=True)
        self.assertEqual(sorted(result["commands"]), sorted(commands))

    def test_iosxr_hostname_deleted(self):
        self.maxDiff = None
        run_cfg = dedent(
            """\
                hostname iosxr1
            """
        )
        self.get_config.return_value = run_cfg
        set_module_args(dict(state="deleted"))
        commands = ["no hostname iosxr1"]
        result = self.execute_module(changed=True)
        self.assertEqual(sorted(result["commands"]), sorted(commands))

    def test_iosxr_hostname_rendered(self):
        self.maxDiff = None
        set_module_args(dict(config=dict(hostname="iosxr1"), state="rendered"))
        commands = ["hostname iosxr1"]
        result = self.execute_module(changed=False)
        self.assertEqual(sorted(result["rendered"]), sorted(commands))

    def test_iosxr_hostname_gathered(self):
        self.maxDiff = None
        run_cfg = dedent(
            """\
                hostname iosxr1
            """
        )
        self.get_config.return_value = run_cfg
        print(self.get_config.return_value)
        set_module_args(dict(state="gathered"))
        gathered = {"hostname": "iosxr1"}
        result = self.execute_module(changed=False)
        self.assertEqual(gathered, result["gathered"])

    def test_iosxr_hostname_parsed(self):
        self.maxDiff = None
        set_module_args(dict(running_config="hostname iosxr1", state="parsed"))
        result = self.execute_module(changed=False)
        parsed_list = {"hostname": "iosxr1"}
        self.assertEqual(parsed_list, result["parsed"])



# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# Copyright (c), Michael DeHaan <michael.dehaan@gmail.com>, 2012-2013
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.module_utils._text import to_native
from ansible.module_utils._text import to_text
from ansible.module_utils.six import string_types


def boto3_tag_list_to_ansible_dict(tags_list, tag_name_key_name=None, tag_value_key_name=None):

    """ Convert a boto3 list of resource tags to a flat dict of key:value pairs
    Args:
        tags_list (list): List of dicts representing AWS tags.
        tag_name_key_name (str): Value to use as the key for all tag keys (useful because boto3 doesn't always use "Key")
        tag_value_key_name (str): Value to use as the key for all tag values (useful because boto3 doesn't always use "Value")
    Basic Usage:
        >>> tags_list = [{'Key': 'MyTagKey', 'Value': 'MyTagValue'}]
        >>> boto3_tag_list_to_ansible_dict(tags_list)
        [
            {
                'Key': 'MyTagKey',
                'Value': 'MyTagValue'
            }
        ]
    Returns:
        Dict: Dict of key:value pairs representing AWS tags
         {
            'MyTagKey': 'MyTagValue',
        }
    """

    if tag_name_key_name and tag_value_key_name:
        tag_candidates = {tag_name_key_name: tag_value_key_name}
    else:
        tag_candidates = {'key': 'value', 'Key': 'Value'}

    # minio seems to return [{}] as an empty tags_list
    if not tags_list or not any(tag for tag in tags_list):
        return {}
    for k, v in tag_candidates.items():
        if k in tags_list[0] and v in tags_list[0]:
            return dict((tag[k], tag[v]) for tag in tags_list)
    raise ValueError("Couldn't find tag key (candidates %s) in tag list %s" % (str(tag_candidates), str(tags_list)))


def ansible_dict_to_boto3_tag_list(tags_dict, tag_name_key_name='Key', tag_value_key_name='Value'):

    """ Convert a flat dict of key:value pairs representing AWS resource tags to a boto3 list of dicts
    Args:
        tags_dict (dict): Dict representing AWS resource tags.
        tag_name_key_name (str): Value to use as the key for all tag keys (useful because boto3 doesn't always use "Key")
        tag_value_key_name (str): Value to use as the key for all tag values (useful because boto3 doesn't always use "Value")
    Basic Usage:
        >>> tags_dict = {'MyTagKey': 'MyTagValue'}
        >>> ansible_dict_to_boto3_tag_list(tags_dict)
        {
            'MyTagKey': 'MyTagValue'
        }
    Returns:
        List: List of dicts containing tag keys and values
        [
            {
                'Key': 'MyTagKey',
                'Value': 'MyTagValue'
            }
        ]
    """

    if not tags_dict:
        return []

    tags_list = []
    for k, v in tags_dict.items():
        tags_list.append({tag_name_key_name: k, tag_value_key_name: to_native(v)})

    return tags_list


def boto3_tag_specifications(tags_dict, types=None):
    """ Converts a list of resource types and a flat dictionary of key:value pairs representing AWS
    resource tags to a TagSpecification object.

    https://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_TagSpecification.html

    Args:
        tags_dict (dict): Dict representing AWS resource tags.
        types (list) A list of resource types to be tagged.
    Basic Usage:
        >>> tags_dict = {'MyTagKey': 'MyTagValue'}
        >>> boto3_tag_specifications(tags_dict, ['instance'])
        [
            {
                'ResourceType': 'instance',
                'Tags': [
                    {
                        'Key': 'MyTagKey',
                        'Value': 'MyTagValue'
                    }
                ]
            }
        ]
    Returns:
        List: List of dictionaries representing an AWS Tag Specification
    """
    if not tags_dict:
        return None
    specifications = list()
    tag_list = ansible_dict_to_boto3_tag_list(tags_dict)

    if not types:
        specifications.append(dict(Tags=tag_list))
        return specifications

    if isinstance(types, string_types):
        types = [types]

    for type_name in types:
        specifications.append(dict(ResourceType=type_name, Tags=tag_list))

    return specifications


def compare_aws_tags(current_tags_dict, new_tags_dict, purge_tags=True):
    """
    Compare two dicts of AWS tags. Dicts are expected to of been created using 'boto3_tag_list_to_ansible_dict' helper function.
    Two dicts are returned - the first is tags to be set, the second is any tags to remove. Since the AWS APIs differ
    these may not be able to be used out of the box.

    :param current_tags_dict:
    :param new_tags_dict:
    :param purge_tags:
    :return: tag_key_value_pairs_to_set: a dict of key value pairs that need to be set in AWS. If all tags are identical this dict will be empty
    :return: tag_keys_to_unset: a list of key names (type str) that need to be unset in AWS. If no tags need to be unset this list will be empty
    """

    tag_key_value_pairs_to_set = {}
    tag_keys_to_unset = []

    for key in current_tags_dict.keys():
        if key not in new_tags_dict and purge_tags:
            tag_keys_to_unset.append(key)

    for key in set(new_tags_dict.keys()) - set(tag_keys_to_unset):
        if to_text(new_tags_dict[key]) != current_tags_dict.get(key):
            tag_key_value_pairs_to_set[key] = new_tags_dict[key]

    return tag_key_value_pairs_to_set, tag_keys_to_unset
